import scipy.stats as ss
import numpy as np
#import statsmodels.api as sm

### Tests on residuals


def normal_Kolmogorov_Smirnov(sample):
    """The moon illumination expressed as a percentage.

                :param astropy sun: the sun ephemeris
                :param astropy moon: the moon ephemeris

                :return: a numpy array like indicated the moon illumination.

                :rtype: array_like

    """

    mu, sigma = ss.norm.fit(sample)
    #use mu sigma for anomaly, 0,1 for rescaling???
    KS_stat, KS_pvalue = ss.kstest(sample, 'norm', args=(0, 1))

    # the sample is likely Gaussian-like if KS_stat (~ maximum distance between sample and theoritical distribution) -> 0
    # the null hypothesis can not be rejected ( i.e the distribution of sample come from a Gaussian) if KS_pvalue -> 1

    KS_judgement = 0

    if KS_pvalue > 0.01:

        KS_judgement = 1

    if KS_pvalue > 0.05:

        KS_judgement = 2

    return KS_stat, KS_pvalue, KS_judgement

def normal_Anderson_Darling(sample):
    """Compute a Anderson-Darling test on the sample versus a normal distribution with mu = 0, sigma = 1

            :param array_like sample: the sample you want to check the "Gaussianity"
            :returns: the Anderson-Darling statistic, the Anderson-Darling critical values associated to the significance
            level of 15 % and the Anderson-Darling judgement
            :rtype: float, array_like, array_like
    """

    AD_stat, AD_critical_values, AD_significance_levels = ss.anderson(sample)

    # the sample is likely Gaussian-like if AD_stat (~ maximum distance between sample and theoritical distribution) -> 0
    # the null hypothesis can not be rejected ( i.e the distribution of sample come from a Gaussian) if AD_pvalue -> 1


    AD_judgement = 0

    if AD_stat < 2*AD_critical_values[-1]:
        AD_judgement = 1

    if AD_stat < AD_critical_values[-1]:
        AD_judgement = 2
    return  AD_stat, AD_critical_values[-1], AD_judgement

def normal_Shapiro_Wilk(sample):
    """Compute a Shapiro-Wilk test on the sample versus a normal distribution with mu = 0, sigma = 1

            :param array_like sample: the sample you want to check the "Gaussianity"
            :returns: the Shapiro-Wilk statistic and its related p_value
            :rtype: float, float
    """

    SW_stat, SW_pvalue = ss.shapiro(sample)

    # the null hypothesis can not be rejected ( i.e the distribution of sample come from a Gaussian) if SW_stat -> 1
    # the null hypothesis can not be rejected ( i.e the distribution of sample come from a Gaussian) if SW_pvalue -> 1

    # Judegement made on the STATISTIC because 'W test statistic is accurate but the p-value may not be" (see scipy doc)
    SW_judgement = 0

    if SW_pvalue > 0.01:
        SW_judgement = 1

    if SW_pvalue > 0.05:
        SW_judgement = 2


    return  SW_stat, SW_pvalue, SW_judgement


### Statistics fit quality metrics

def normalized_chi2(chi2, n_data, n_parameters) :
    """Compute the chi^2/dof

            :param float chi2: the chi^2
            :param int n_data: the number of data_points
            :param int n_parameters: the number of model parameters

            :returns: the chi^2/dof and the chi2dof_judgement
            :rtype: float
    """

    chi2_sur_dof = chi2/(n_data-n_parameters)

    chi2dof_judgement = 0
    if chi2_sur_dof < 2 :
        chi2dof_judgement = 2

    return chi2_sur_dof,chi2dof_judgement

def Bayesian_Information_Criterion(chi2, n_data, n_parameters):
    """Compute the BIC statistic.

            :param float chi2: the chi^2
            :param int n_data: the number of data_points
            :param int n_parameters: the number of model parameters

            :returns: the chi^2/dof
            :rtype: float
    """
    BIC = chi2 + n_parameters*np.log(n_data)

    return BIC

def Akaike_Information_Criterion(chi2, n_parameters):
    """Compute the BIC statistic.

            :param float chi2: the chi^2
            :param int n_parameters: the number of model parameters

            :returns: the chi^2/dof
            :rtype: float
    """
    AIC = chi2 + 2*n_parameters

    return AIC





"""
Created on Thu Apr 20 11:19:38 2017

@author: ebachelet
"""
from __future__ import division

import numpy as np
import time
import scipy.spatial as ss
import matplotlib.pyplot as plt


def find_2_lenses_caustics_and_critical_curves(separation, mass_ratio, resolution=1000):
    """  Find and sort caustics for a binary lens

        :param float separation: the projected normalised angular distance between the two bodies
        :param float mass_ratio: the mass ratio of the two bodies
        :param int resolution: number of points desired in the caustic computation.

        :return: the caustic regime,  the sorted caustics and the critical curve
        :rtype: str, list of array_like, list of array_like
    """
    close_top_caustic = None
    close_top_cc = None
    close_bottom_caustic = None
    close_bottom_cc = None
    central_caustic = None
    central_cc = None
    wide_caustic = None
    wide_cc = None
    resonant_caustic = None
    resonant_cc = None

    caustics_points, critical_curve_points = compute_2_lenses_caustics_points(separation, mass_ratio,
                                                                              resolution=resolution)

    critical_curve = critical_curve_points

    caustic_regime = find_2_lenses_caustic_regime(separation, mass_ratio)

    if caustic_regime == 'resonant':
        resonant_caustic, resonant_cc = sort_2lenses_resonant_caustic(caustics_points, critical_curve_points)

    if caustic_regime == 'close':
        result = sort_2lenses_close_caustics(caustics_points, critical_curve_points)
        central_caustic, close_top_caustic, close_bottom_caustic, central_cc, close_top_cc, close_bottom_cc = result

    if caustic_regime == 'wide':
        central_caustic, wide_caustic, central_cc, wide_cc = sort_2lenses_wide_caustics(caustics_points,
                                                                                        critical_curve_points)

    caustics = [central_caustic, close_top_caustic, close_bottom_caustic, wide_caustic, resonant_caustic]
    critical_curve = [central_cc, close_top_cc, close_bottom_cc, wide_cc, resonant_cc]

    return caustic_regime, caustics, critical_curve


def sort_2lenses_resonant_caustic(caustic_points, critical_curves_points):
    """ Sort the caustic points to have regular resonant caustic

        :param array_like caustic_points: the points of the caustics, in complex notation
        :param array_like critical_curves_points: the points of the critical curve, in complex notation

        :return: the resonant caustic,  the resonant critical curve
        :rtype:  array_like, array_like
    """
    try:
        medians_y = np.median(caustic_points[:, :].imag, axis=0)

        positive_y_branches = np.where(medians_y > 0)[0]

        first_branch = positive_y_branches[0]
        second_branch = positive_y_branches[1]

        if np.max((caustic_points[:, first_branch]).real) > np.max((caustic_points[:, second_branch]).real):

            resonant_caustic = np.r_[caustic_points[:, second_branch], caustic_points[:, first_branch],
                                     np.conj(caustic_points[:, first_branch][::-1]),
                                     np.conj(caustic_points[:, second_branch][::-1])]

            resonant_cc = np.r_[critical_curves_points[:, second_branch], critical_curves_points[:, first_branch],
                                np.conj(critical_curves_points[:, first_branch][::-1]),
                                np.conj(critical_curves_points[:, second_branch][::-1])]
        else:

            resonant_caustic = np.r_[caustic_points[:, first_branch], caustic_points[:, second_branch],
                                     np.conj(caustic_points[:, second_branch][::-1]),
                                     np.conj(caustic_points[:, first_branch][::-1])]

            resonant_cc = np.r_[critical_curves_points[:, first_branch], critical_curves_points[:, second_branch],
                                np.conj(critical_curves_points[:, second_branch][::-1]),
                                np.conj(critical_curves_points[:, first_branch][::-1])]

    except:
        import pdb;
        pdb.set_trace()
        resonant_caustic = caustic_points

        resonant_cc = critical_curves_points

    return resonant_caustic, resonant_cc


def sort_2lenses_close_caustics(caustic_points, critical_curves_points):
    """ Sort the caustic points to have one central caustic and two "planetary" caustics

        :param array_like caustic_points: the points of the caustics, in complex notation
        :param array_like critical_curves_points: the points of the critical curve, in complex notation

        :return: the central caustic,  the top planetary caustic, the bottom planetary caustic,
        the central critical curve, the top critical curve, the bottom critical curve
        :rtype:  array_like, array_like, array_like, array_like, array_like, array_like
    """

    try:
        medians_y = np.median(caustic_points[:, :].imag, axis=0)

        order = np.argsort(medians_y)

        close_bottom_caustic = caustic_points[:, order[0]]
        close_bottom_cc = critical_curves_points[:, order[0]]

        close_top_caustic = caustic_points[:, order[-1]]
        close_top_cc = critical_curves_points[:, order[-1]]

        if np.abs(caustic_points[-1, order[1]].real - caustic_points[0, order[2]].real) < np.abs(
            caustic_points[-1, order[1]].real - caustic_points[:, order[2]][::-1][0].real):

            central_caustic = np.r_[caustic_points[:, order[1]], caustic_points[:, order[2]]]
            central_cc = np.r_[critical_curves_points[:, order[1]], critical_curves_points[:, order[2]]]

        else:

            central_caustic = np.r_[caustic_points[:, order[1]], caustic_points[:, order[2]][::-1]]
            central_cc = np.r_[critical_curves_points[:, order[1]], critical_curves_points[:, order[2]][::-1]]


    except:

        medians_y = np.median(caustic_points[:, :].imag)

        order = np.argsort(medians_y)

        close_bottom_caustic = caustic_points[:, order[0]]
        close_bottom_cc = critical_curves_points[:, order[0]]

        close_top_caustic = caustic_points[:, order[-1]]
        close_top_cc = critical_curves_points[:, order[-1]]

        if np.abs(caustic_points[-1, order[1]].real - caustic_points[0, order[2]].real) < np.abs(
                        caustic_points[-1, order[1]].real - caustic_points[:, order[2]][::-1][0].real):

            central_caustic = np.r_[caustic_points[:, order[1]], caustic_points[:, order[2]]]
            central_cc = np.r_[critical_curves_points[:, order[1]], critical_curves_points[:, order[2]]]

        else:

            central_caustic = np.r_[caustic_points[:, order[1]], caustic_points[:, order[2]][::-1]]
            central_cc = np.r_[critical_curves_points[:, order[1]], critical_curves_points[:, order[2]][::-1]]


    return central_caustic, close_top_caustic, close_bottom_caustic, central_cc, close_top_cc, close_bottom_cc


def sort_2lenses_wide_caustics(caustic_points, critical_curves_points):
    """ Sort the caustic points to have one central caustic and one "planetary" caustics

        :param array_like caustic_points: the points of the caustics, in complex notation
        :param array_like critical_curves_points: the points of the critical curve, in complex notation

        :return: the central caustic,  the  planetary caustic, the central critical curve,
        the planetary critical curve
        :rtype:  array_like, array_like, array_like, array_like
    """
    try:
        medians_y = np.median(caustic_points[:, :].imag, axis=0)

        order = np.argsort(medians_y)

        wide_bottom_caustic = caustic_points[:, order[0]]
        wide_bottom_cc = critical_curves_points[:, order[0]]

        central_bottom_caustic = caustic_points[:, order[1]]
        central_bottom_cc = critical_curves_points[:, order[1]]

        wide_top_caustic = caustic_points[:, order[-1]]
        wide_top_cc = critical_curves_points[:, order[-1]]

        central_top_caustic = caustic_points[:, order[2]]
        central_top_cc = critical_curves_points[:, order[2]]

        central_caustic = np.r_[central_bottom_caustic, central_top_caustic]
        central_cc = np.r_[central_bottom_cc, central_top_cc]

        wide_caustic = np.r_[wide_bottom_caustic, wide_top_caustic]
        wide_cc = np.r_[wide_bottom_cc, wide_top_cc]

    except:

        medians_y = np.median(caustic_points[:, :].imag)

        order = np.argsort(medians_y)

        wide_bottom_caustic = caustic_points[:, order[0]]
        wide_bottom_cc = critical_curves_points[:, order[0]]

        central_bottom_caustic = caustic_points[:, order[1]]
        central_bottom_cc = critical_curves_points[:, order[1]]

        wide_top_caustic = caustic_points[:, order[-1]]
        wide_top_cc = critical_curves_points[:, order[-1]]

        central_top_caustic = caustic_points[:, order[2]]
        central_top_cc = critical_curves_points[:, order[2]]

        central_caustic = np.r_[central_bottom_caustic, central_top_caustic]
        central_cc = np.r_[central_bottom_cc, central_top_cc]

        wide_caustic = np.r_[wide_bottom_caustic, wide_top_caustic]
        wide_cc = np.r_[wide_bottom_cc, wide_top_cc]

    return central_caustic, wide_caustic, central_cc, wide_cc


def compute_2_lenses_caustics_points(separation, mass_ratio, resolution=1000):
    """  Find the critical curve points and caustics points associated to a binary lens. See :
         "On the Minimum Magnification Between Caustic Crossings for Microlensing by Binary and Multiple Stars"
         Witt and Mao 1995 http://adsabs.harvard.edu/abs/1995ApJ...447L.105W
         "Investigation of high amplification events in light curves of gravitationally lensed quasars"
         Witt H.J 1990 http://adsabs.harvard.edu/abs/1990A%26A...236..311W

        :param float separation: the projected normalised angular distance between the two bodies
        :param float mass_ratio: the mass ratio of the two bodies
        :param int resolution: number of points desired in the caustic computation.
                 :return: the central caustic,  the top planetary caustic, the bottom planetary caustic,
                 the central critical curve, the top critical curve, the bottom critical curve
                 :rtype:  array_like, array_like, array_like, array_like, array_like, array_like
    """
    caustics = []
    critical_curves = []

    center_of_mass = mass_ratio / (1 + mass_ratio) * separation

    # Witt&Mao magic numbers
    total_mass = 0.5
    mass_1 = 1 / (1 + mass_ratio)
    mass_2 = mass_ratio * mass_1
    delta_mass = (mass_2 - mass_1) / 2
    lens_1 = -separation / 2.0
    lens_2 = separation / 2.0
    lens_1_conjugate = np.conj(lens_1)
    lens_2_conjugate = np.conj(lens_2)

    phi = np.linspace(0.00, 2 * np.pi, resolution)
    roots = []
    wm_1 = 4.0 * lens_1 * delta_mass
    wm_3 = 0.0
    slice_1 = []
    slice_2 = []
    slice_3 = []
    slice_4 = []
    slices = [slice_1, slice_2, slice_3, slice_4]
    for angle in phi:

        e_phi = np.cos(-angle) + 1j * np.sin(-angle)  # See Witt & Mao

        wm_0 = -2.0 * total_mass * lens_1 ** 2 + e_phi * lens_1 ** 4
        wm_2 = -2.0 * total_mass - 2 * e_phi * lens_1 ** 2
        wm_4 = e_phi

        polynomial_coefficients = [wm_4, wm_3, wm_2, wm_1, wm_0]

        polynomial_roots = np.roots(polynomial_coefficients)

        checks = np.polyval(polynomial_coefficients, polynomial_roots)

        if np.max(np.abs(checks)) > 10 ** -10:
            pass
        else:
            # polynomial_roots = np.polynomial.polynomial.polyroots(polynomial_coefficients[::-1])

            if len(roots) == 0:

                pol_roots = polynomial_roots

            else:

                aa = np.c_[polynomial_roots.real, polynomial_roots.imag]
                bb = np.c_[roots[-1].real, roots[-1].imag]

                distances = ss.distance.cdist(aa, bb)
                good_order = [0, 0, 0, 0]

                for i in range(4):
                    line, column = np.where((distances) == np.min(distances))
                    good_order[column[0]] = polynomial_roots[line[0]]
                    distances[line[0], :] += 10 ** 10
                    distances[:, column[0]] += 10 ** 10

                pol_roots = np.array(good_order)

            roots.append(pol_roots)

            images_conjugate = np.conj(pol_roots)
            zeta_caustics = pol_roots + mass_1 / (lens_1_conjugate - images_conjugate) + mass_2 / (
                lens_2_conjugate - images_conjugate)

            if len(caustics) == 0:
                caustics = zeta_caustics
                critical_curves = pol_roots
            else:
                caustics = np.vstack((caustics, zeta_caustics))
                critical_curves = np.vstack((critical_curves, pol_roots))

    # shift into center of mass referentiel

    caustics += -center_of_mass + separation / 2
    critical_curves += -center_of_mass + separation / 2

    return caustics, critical_curves


def find_area_of_interest_around_caustics(caustics, secure_factor=0.1):
    """  Find a box around the caustic, the are of interest.

       :param list caustics: a list containing all array_like of caustics
       :param float secure_factor: an extra limit around the box, in Einstein ring unit

       :return: the are of interest,  a list containin x limits, y limits and x,y center
       :rtype: list
   """
    all_points = []
    for caustic in caustics:

        if caustic is not None:
            all_points += caustic.ravel().tolist()
    all_points = np.array(all_points)
    min_X = np.min(all_points.real) - secure_factor
    max_X = np.max(all_points.real) + secure_factor

    min_Y = np.min(all_points.imag) - secure_factor
    max_Y = np.max(all_points.imag) + secure_factor

    center_of_the_box_X = min_X + (max_X - min_X) / 2
    center_of_the_box_Y = min_Y + (max_Y - min_Y) / 2

    area_of_interest = [[min_X, max_X], [min_Y, max_Y], [center_of_the_box_X, center_of_the_box_Y]]

    return area_of_interest


def find_2_lenses_caustic_regime(separation, mass_ratio):
    """  Find the caustic regime.
        "An alternative parameterisation for binary-lens caustic-crossing events"
        Cassan A.2008 http://adsabs.harvard.edu/abs/2008A%26A...491..587C
        "Speeding up Low-mass Planetary Microlensing Simulations and Modeling: The Caustic Region Of INfluence"
        Penny M.2014 http://adsabs.harvard.edu/abs/2014ApJ...790..142P

       :param float separation: the projected normalised angular distance between the two bodies
       :param float mass_ratio: the mass ratio of the two bodies

       :return: caustic_regime: close, wide or resonant
       :rtype: str
   """

    caustic_regime = 'resonant'

    wide_limit = ((1 + mass_ratio ** (1 / 3.0)) ** 3.0 / (1 + mass_ratio)) ** 0.5

    if (separation > 1.0) & (separation > wide_limit):
        caustic_regime = 'wide'
        return caustic_regime

    close_limit = mass_ratio / (1 + mass_ratio) ** 2.0

    if (separation < 1.0) & (1 / separation ** 8.0 * ((1 - separation ** 4.0) / 3.0) ** 3.0 > close_limit):
        caustic_regime = 'close'
        return caustic_regime

    return caustic_regime


def change_source_trajectory_center_to_central_caustics_center(separation, mass_ratio):
    """  Define a new geometric center, on the central caustic
            resonant : (0,0)
            close : (x_central_caustic,0)
            wide : (x_central_caustic, 0)

        "Speeding up Low-mass Planetary Microlensing Simulations and Modeling: The Caustic Region Of INfluence"
        Penny M.2014 http://adsabs.harvard.edu/abs/2014ApJ...790..142P
        "Microlensing observations rapid search for exoplanets: MORSE code for GPUs"
        McDougall A. and Albrow M. http://adsabs.harvard.edu/abs/2016MNRAS.456..565M

       :param float separation: the projected normalised angular distance between the two bodies
       :param float mass_ratio: the mass ratio of the two bodies


       :return: x_center, y_center : the new system center
       :rtype: str

   """
    x_center = 0
    y_center = 0

    caustic_regime, caustics, critical_curves = find_2_lenses_caustics_and_critical_curves(separation, mass_ratio,
                                                                                           resolution=10)
    # caustic_regime = find_2_lenses_caustic_regime(separation, mass_ratio)

    if caustic_regime == 'wide':
        x_center = np.median(caustics[0].real)


        # pass
    if caustic_regime == 'close':
        x_center = np.median(caustics[0].real)
        y_center = 0

    return x_center, y_center


def change_source_trajectory_center_to_planetary_caustics_center(separation, mass_ratio):
    """  Define a new geometric center depending of the caustic regime. This helps fit convergence sometimes:
            resonant : (0,0)
            close : (x_planetary_caustic, y_planetary_caustic)
            wide : (x_planetary_caustic, y_planetary_caustic)

        "Speeding up Low-mass Planetary Microlensing Simulations and Modeling: The Caustic Region Of INfluence"
        Penny M.2014 http://adsabs.harvard.edu/abs/2014ApJ...790..142P
        "Microlensing observations rapid search for exoplanets: MORSE code for GPUs"
        McDougall A. and Albrow M. http://adsabs.harvard.edu/abs/2016MNRAS.456..565M

       :param float separation: the projected normalised angular distance between the two bodies
       :param float mass_ratio: the mass ratio of the two bodies


       :return: x_center, y_center : the new system center
       :rtype: str

   """
    x_center = 0
    y_center = 0

    caustic_regime, caustics, critical_curves = find_2_lenses_caustics_and_critical_curves(separation, mass_ratio,
                                                                                           resolution=10)

    if caustic_regime == 'wide':
        x_center = np.median(caustics[-2].real)

    if caustic_regime == 'close':
        x_center = np.median(caustics[1].real)
        y_center = np.median(caustics[1].imag)

    return x_center, y_center


### DEPRECIATED
def dx(function, x, args):
    return abs(0 - function(x, args[0], args[1], args[2], args[3], args[4]))


def newtons_method(function, jaco, x0, args=None, e=10 ** -10):
    delta = dx(function, x0, args)

    while np.max(delta) > e:
        x0 = x0 - function(x0, args[0], args[1], args[2], args[3], args[4]) / jaco(x0, args[0], args[1], args[2],
                                                                                   args[3], args[4])
        delta = dx(function, x0, args)
    return x0


def foo(x, pp1, pp2, pp3, pp4, pp5):
    x_2 = x ** 2
    x_3 = x_2 * x
    x_4 = x_2 * x_2
    aa = pp1 * x_4 + pp2 * x_3 + pp3 * x_2 + pp4 * x + pp5

    return aa


def foo2(X, pp1, pp2, pp3, pp4, pp5):
    x = X[0] + 1j * X[1]

    x_2 = x ** 2
    x_3 = x_2 * x
    x_4 = x_2 * x_2
    aa = pp1 * x_4 + pp2 * x_3 + pp3 * x_2 + pp4 * x + pp5

    return [aa.real, aa.imag]


def jaco(x, pp1, pp2, pp3, pp4, pp5):
    x_2 = x ** 2
    x_3 = x_2 * x
    aa = 4 * pp1 * x_3 + 3 * pp2 * x_2 + 2 * pp3 * x + pp4

    return aa


def hessian(x, pp1, pp2, pp3, pp4, pp5):
    aa = 12 * pp1 * x ** 2 + 6 * pp2 * x + 2 * pp3

    return aa


def Jacobian(zi, mtot, deltam, z1):
    ziconj = np.conj(zi)

    dzeta = (mtot - deltam) / (np.conjugate(z1) - ziconj) ** 2 + (mtot + deltam) / (-np.conjugate(z1) - ziconj) ** 2
    dzetaconj = np.conj(dzeta)

    dZETA = dzeta * dzetaconj

    detJ = 1 - dZETA.real
    return np.sign(detJ)


def find_slices(slice_1, slice_2, points):
    centroid_1 = np.median(slice_1.real) + 1j * np.median(slice_1.imag)
    centroid_2 = np.median(slice_2.real) + 1j * np.median(slice_2.imag)



# -*- coding: utf-8 -*-
"""
Created on Mon Dec  7 10:32:13 2015

@author: ebachelet
"""

from __future__ import division
from collections import OrderedDict
import os.path
import abc
import sys
from scipy import interpolate, misc
import pkg_resources

thismodule = sys.modules[__name__]

import numpy as np
import collections
import time as python_time

from pyLIMA import microlguess
from pyLIMA import microlmagnification
from pyLIMA import microlpriors
from pyLIMA import microlparallax
from pyLIMA import microlorbitalmotion
from pyLIMA import microlcaustics
from pyLIMA import stars
from pyLIMA import microlxallarap

resource_package = __name__
resource_path = '/'.join(('data', 'Yoo_B0B1.dat'))
template = pkg_resources.resource_filename(resource_package, resource_path)
try:

    # yoo_table = np.loadtxt('b0b1.dat')
    yoo_table = np.loadtxt(template)
except:

    print('ERROR : No Yoo_B0B1.dat file found, please check!')

b0b1 = yoo_table
zz = b0b1[:, 0]
b0 = b0b1[:, 1]
b1 = b0b1[:, 2]
# db0 = b0b1[:,4]
# db1 = b0b1[:, 5]
interpol_b0 = interpolate.interp1d(zz, b0, kind='linear')
interpol_b1 = interpolate.interp1d(zz, b1, kind='linear')
# import pdb; pdb.set_trace()

dB0 = misc.derivative(lambda x: interpol_b0(x), zz[1:-1], dx=10 ** -4, order=3)
dB1 = misc.derivative(lambda x: interpol_b1(x), zz[1:-1], dx=10 ** -4, order=3)
dB0 = np.append(2.0, dB0)
dB0 = np.concatenate([dB0, [dB0[-1]]])
dB1 = np.append((2.0 - 3 * np.pi / 4), dB1)
dB1 = np.concatenate([dB1, [dB1[-1]]])
interpol_db0 = interpolate.interp1d(zz, dB0, kind='linear')
interpol_db1 = interpolate.interp1d(zz, dB1, kind='linear')
yoo_table = [zz, interpol_b0, interpol_b1, interpol_db0, interpol_db1]


class ModelException(Exception):
    pass


def create_model(model_type, event, model_arguments=[], parallax=['None', 0.0], xallarap='None',
                 orbital_motion=['None', 0.0], source_spots='None', blend_flux_ratio=True):
    """
    Load a model according to the supplied model_type. Models are expected to be named
    Model<model_type> e.g. ModelPSPL

    :param string model_type: Model type e.g. PSPL
    :return: Model object for given model_type
    """

    try:

        model = getattr(thismodule, 'Model{}'.format(model_type))

    except AttributeError:

        raise ModelException('Unknown model "{}"'.format(model_type))

    return model(event, model_arguments, parallax, xallarap,
                 orbital_motion, source_spots, blend_flux_ratio)


class MLModel(object):
    """
       ######## MLModels module ########

       This class defines the model you want to fit your data to. Model is the parent class, each model is a child
       class (polymorphism), for example ModelPSPL.

       Attributes :

           event : A event class which describe your event that you want to model. See the event module.


           parallax_model : Parallax model you want to use for the Earth types telescopes.
                      Has to be a list containing the model in the available_parallax
                      parameter and the value of topar. Have a look here for more details :
                      http://adsabs.harvard.edu/abs/2011ApJ...738...87S

                       'Annual' --> Annual parallax
                       'Terrestrial' --> Terrestrial parallax
                       'Full' --> combination of previous

                       topar --> a time in HJD choosed as the referenced time fot the parallax

                     If you have some Spacecraft types telescopes, the space based parallax
                     is computed if parallax is different of 'None'
                     More details in the microlparallax module

           xallarap_model : not available yet

           orbital_motion_model : not available yet

                   'None' --> No orbital motion
                   '2D' --> Classical orbital motion
                   '3D' --> Full Keplerian orbital motion

                   toom --> a time in HJD choosed as the referenced time fot the orbital motion
                           (Often choose equal to topar)

                   More details in the microlomotion module

           source_spots_model : not available yet

                   'None' --> No source spots

                    More details in the microlsspots module

            yoo_table : an array which contains the Yoo et al table

            Jacobian_flag : a flag indicated if a Jacobian can be used ('OK') or not.

            model_dictionnary : a python dictionnary which describe the model parameters

            pyLIMA_standards_dictionnary : the standard pyLIMA parameters dictionnary

            fancy_to_pyLIMA_dictionnary : a dictionnary which described which fancy parameters replace a standard pyLIMA
             parameter. For example : {'logrho': 'rho'}

            pyLIMA_to_fancy : a dictionnary which described the function to transform the standard pyLIMA parameter to
            the fancy one. Example :  {'logrho': lambda parameters: np.log10(parameters.rho)}

            fancy_to_pyLIMA : a dictionnary which described the function to transform the fancy parameters to
            pyLIMA standards. Example :  {'rho': lambda parameters: 10 ** parameters.logrho}

            parameters_guess : a list containing guess on pyLIMA parameters.


       :param object event: a event object. More details on the event module.
       :param list parallax: a list of [string,float] indicating the parallax model you want and to_par
       :param list xallarap: a list of [string,float] indicating the xallarap mode.l. NOT WORKING NOW.
       :param list orbital_motion: a list of [string,float] indicating the parallax model you want and to_om.
                                   NOT WORKING NOW.
       :param string source_spots: a string indicated the source_spots you want. NOT WORKING.
       """
    __metaclass__ = abc.ABCMeta

    def __init__(self, event, model_arguments=[], parallax=['None', 0.0], xallarap='None',
                 orbital_motion=['None', 0.0], source_spots='None', blend_flux_ratio=True):
        """ Initialization of the attributes described above.
        """
        self.event = event
        self.model_arguments = model_arguments
        self.parallax_model = parallax
        self.xallarap_model = xallarap
        self.orbital_motion_model = orbital_motion
        self.source_spots_model = source_spots
        self.yoo_table = yoo_table
        self.blend_flux_ratio = blend_flux_ratio

        self.model_dictionnary = {}
        self.pyLIMA_standards_dictionnary = {}

        self.fancy_to_pyLIMA_dictionnary = {}
        self.pyLIMA_to_fancy = {}
        self.fancy_to_pyLIMA = {}

        self.parameters_guess = []
        self.Jacobian_flag = 'OK'

        self.define_pyLIMA_standard_parameters()

        # binary lens model specific
        self.binary_origin = None
        self.x_center = None
        self.y_center = None

    @abc.abstractproperty
    def model_type(self):
        pass

    @abc.abstractmethod
    def paczynski_model_parameters(self):
        return

    @abc.abstractmethod
    def model_magnification(self, telescope, pyLIMA_parameters):
        return

    def define_pyLIMA_standard_parameters(self):
        """ Define the standard pyLIMA parameters dictionnary."""

        self.pyLIMA_standards_dictionnary = self.paczynski_model_parameters()

        if self.parallax_model[0] != 'None':
            self.Jacobian_flag = 'No way'
            self.pyLIMA_standards_dictionnary['piEN'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['piEE'] = len(self.pyLIMA_standards_dictionnary)

            self.event.compute_parallax_all_telescopes(self.parallax_model)

        if self.xallarap_model != 'None':
            self.Jacobian_flag = 'No way'
            self.pyLIMA_standards_dictionnary['XiEN'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['XiEE'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['ra_xallarap'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['dec_xallarap'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['period_xallarap'] = len(self.pyLIMA_standards_dictionnary)
            if self.xallarap_model != 'circular' :
                self.pyLIMA_standards_dictionnary['eccentricity_xallarap'] = len(self.pyLIMA_standards_dictionnary)
                self.pyLIMA_standards_dictionnary['t_periastron_xallarap'] = len(self.pyLIMA_standards_dictionnary)


        if self.orbital_motion_model[0] != 'None':
            self.Jacobian_flag = 'No way'
            self.pyLIMA_standards_dictionnary['dsdt'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['dalphadt'] = len(self.pyLIMA_standards_dictionnary)

        if self.source_spots_model != 'None':
            self.Jacobian_flag = 'No way'
            self.pyLIMA_standards_dictionnary['spot'] = len(self.pyLIMA_standards_dictionnary)

        for telescope in self.event.telescopes:
            self.pyLIMA_standards_dictionnary['fs_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)

            if self.blend_flux_ratio:
                self.pyLIMA_standards_dictionnary['g_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)
            else:
                self.pyLIMA_standards_dictionnary['fb_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)

        self.pyLIMA_standards_dictionnary = OrderedDict(
            sorted(self.pyLIMA_standards_dictionnary.items(), key=lambda x: x[1]))

        self.parameters_boundaries = microlguess.differential_evolution_parameters_boundaries(self)

    def define_model_parameters(self):
        """ Define the model parameters dictionnary. It is different to the pyLIMA_standards_dictionnary
         if you have some fancy parameters request.
        """
        self.model_dictionnary = self.pyLIMA_standards_dictionnary.copy()
        if len(self.pyLIMA_to_fancy) != 0:

            self.Jacobian_flag = 'No way'
            for key_parameter in self.fancy_to_pyLIMA_dictionnary.keys():
                try:
                    self.model_dictionnary[key_parameter] = self.model_dictionnary.pop(
                        self.fancy_to_pyLIMA_dictionnary[key_parameter])
                except:

                    pass
            self.model_dictionnary = OrderedDict(
                sorted(self.model_dictionnary.items(), key=lambda x: x[1]))

        self.model_parameters = collections.namedtuple('parameters', self.model_dictionnary)

    def compute_the_microlensing_model(self, telescope, pyLIMA_parameters):
        """ Compute the microlens model according the injected parameters. This is modified by child submodel sublclass,
        if not the default microlensing model is returned.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :returns: the microlensing model
        :rtype: array_like
        """

        amplification = self.model_magnification(telescope, pyLIMA_parameters)
        return self._default_microlensing_model(telescope, pyLIMA_parameters, amplification)

    def _default_microlensing_model(self, telescope, pyLIMA_parameters, amplification):
        """ Compute the default microlens model according the injected parameters:

        flux(t) = f_source*magnification(t)+f_blending

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :param array_like amplification: the magnification associated to the model
        :returns: the microlensing model, the microlensing priors
        :rtype: array_like, float
        """

        f_source, g_blending = self.derive_telescope_flux(telescope, pyLIMA_parameters, amplification)

        if self.blend_flux_ratio:
            microlensing_model = f_source * (amplification + g_blending)
        else:
            microlensing_model = f_source * amplification + g_blending

        return microlensing_model, f_source, g_blending

    def derive_telescope_flux(self, telescope, pyLIMA_parameters, amplification):
        """
        Compute the source/blending flux

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :param array_like amplification: an array containing the magnification

        :returns:  the source and the blending flux
        :rtype: tuple
        """
        try:
            # Fluxes parameters are fitted
            f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope.name) / 2

            if self.blend_flux_ratio:
                g_blending = 2 * getattr(pyLIMA_parameters, 'g_' + telescope.name) / 2
            else:
                g_blending = 2 * getattr(pyLIMA_parameters, 'fb_' + telescope.name) / 2


        except TypeError:

            # Fluxes parameters are estimated through np.polyfit
            lightcurve = telescope.lightcurve_flux
            flux = lightcurve[:, 1]
            errflux = lightcurve[:, 2]

            try:
                f_source, f_blending = np.polyfit(amplification, flux, 1, w=1 / errflux)

                if self.blend_flux_ratio:
                    g_blending = f_blending / f_source
                else:
                    g_blending = f_blending

            except:

                f_source = 0.0
                g_blending = 0.0

        return f_source, g_blending

    def compute_pyLIMA_parameters(self, fancy_parameters):
        """ Realize the transformation between the fancy parameters to fit to the
        standard pyLIMA parameters needed to compute a model.

        :param list fancy_parameters: the parameters you fit
        :return: pyLIMA parameters
        :rtype: object (namedtuple)
        """
        # start_time = python_time.time()


        for key_parameter in self.model_dictionnary.keys():

            try:

                setattr(self.model_parameters, key_parameter, fancy_parameters[self.model_dictionnary[key_parameter]])

            except:
                setattr(self.model_parameters, key_parameter, None)

        # print 'arange', python_time.time() - start_time

        pyLIMA_parameters = self.fancy_parameters_to_pyLIMA_standard_parameters(self.model_parameters)

        # print 'conversion', python_time.time() - start_time
        return pyLIMA_parameters

    def fancy_parameters_to_pyLIMA_standard_parameters(self, fancy_parameters):
        """ Transform the fancy parameters to the pyLIMA standards. The output got all
        the necessary standard attributes, example to, uo, tE...


        :param object fancy_parameters: the fancy_parameters as namedtuple
        :return: the pyLIMA standards are added to the fancy parameters
        :rtype: object
        """
        # start_time = python_time.time()
        if len(self.fancy_to_pyLIMA) != 0:

            for key_parameter in self.fancy_to_pyLIMA.keys():
                setattr(fancy_parameters, key_parameter, self.fancy_to_pyLIMA[key_parameter](fancy_parameters))

        # print 'fancy to PYLIMA', python_time.time() - start_time
        return fancy_parameters

    def pyLIMA_standard_parameters_to_fancy_parameters(self, pyLIMA_parameters):
        """ Transform the  the pyLIMA standards parameters to the fancy parameters. The output got all
            the necessary fancy attributes.


        :param object pyLIMA_parameters: the  standard pyLIMA parameters as namedtuple
        :return: the fancy parameters are added to the fancy parameters
        :rtype: object
        """
        if len(self.pyLIMA_to_fancy) != 0:

            for key_parameter in self.pyLIMA_to_fancy.keys():
                setattr(pyLIMA_parameters, key_parameter, self.pyLIMA_to_fancy[key_parameter](pyLIMA_parameters))

        return pyLIMA_parameters

    def source_trajectory(self, telescope, to, uo, tE, pyLIMA_parameters):
        """ Compute the microlensing source trajectory associated to a telescope for the given parameters.

        :param float to: time of maximum magnification
        :param float uo: minimum impact parameter
        :param float tE: angular Einstein ring crossing time
        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: source_trajectory_x, source_trajectory_y the x,y compenents of the source trajectory
        :rtype: array_like,array_like
        """
        # Linear basic trajectory

        lightcurve = telescope.lightcurve_flux
        time = lightcurve[:, 0]

        tau = (time - to) / tE
        beta = np.array([uo] * len(tau))

        # These following second order induce curvatures in the source trajectory
        # Parallax?
        if 'piEN' in pyLIMA_parameters._fields:
            piE = np.array([pyLIMA_parameters.piEN, pyLIMA_parameters.piEE])
            parallax_delta_tau, parallax_delta_beta = microlparallax.compute_parallax_curvature(piE,
                                                                                                telescope.deltas_positions)

            tau += parallax_delta_tau
            beta += parallax_delta_beta

        # Xallarap?
        if 'XiEN' in pyLIMA_parameters._fields:
            XiE = np.array([pyLIMA_parameters.XiEN, pyLIMA_parameters.XiEE])
            ra = pyLIMA_parameters.ra_xallarap
            dec = pyLIMA_parameters.dec_xallarap
            period = pyLIMA_parameters.period_xallarap
            if 'eccentricity_xallarap' in pyLIMA_parameters._fields:
                eccentricity = pyLIMA_parameters.eccentricity_xallarap
                t_periastron = pyLIMA_parameters.t_periastron_xallarap

                orbital_elements = [telescope.lightcurve_flux[:,0], ra,dec,period,eccentricity,t_periastron]
                xallarap_delta_tau, xallarap_delta_beta = microlxallarap.compute_xallarap_curvature(XiE, orbital_elements,
                                                                                                    mode ='elliptic')
            else:

                orbital_elements = [telescope.lightcurve_flux[:, 0], ra, dec, period]
                xallarap_delta_tau, xallarap_delta_beta = microlxallarap.compute_xallarap_curvature(XiE,
                                                                                                    orbital_elements)

            tau += xallarap_delta_tau
            beta += xallarap_delta_beta
        # Orbital motion?
        if 'alpha' in pyLIMA_parameters._fields:

            alpha = pyLIMA_parameters.alpha
            if 'dalphadt' in pyLIMA_parameters._fields:
                alpha += microlorbitalmotion.orbital_motion_2D_trajectory_shift(self.orbital_motion_model[1],
                                                                                telescope.lightcurve_flux[:, 0],
                                                                                pyLIMA_parameters.dalphadt)
            source_trajectory_x = tau * np.cos(alpha) - beta * np.sin(alpha)
            source_trajectory_y = tau * np.sin(alpha) + beta * np.cos(alpha)

        else:

            source_trajectory_x = tau
            source_trajectory_y = beta

        return source_trajectory_x, source_trajectory_y


class ModelPSPL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns: PSPL
        :rtype: string
        """
        return 'PSPL'

    def paczynski_model_parameters(self):
        """ Define the PSPL standard parameters, [to,uo,tE]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2}

        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a PSPL model. More details in microlmagnification module.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification
        :rtype: array_like
        """

        source_trajectoire = self.source_trajectory(telescope, pyLIMA_parameters.to, pyLIMA_parameters.uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        return microlmagnification.amplification_PSPL(*source_trajectoire)

    def Jacobian_model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a PSPL model. More details in microlmagnification module.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification, impact_parameter
        :rtype: array_like,array_like
        """

        source_trajectoire = self.source_trajectory(telescope, pyLIMA_parameters.to, pyLIMA_parameters.uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        return microlmagnification.Jacobian_amplification_PSPL(*source_trajectoire)

    def model_Jacobian(self, telescope, pyLIMA_parameters):
        """ The derivative of a PSPL model

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: jacobi
        :rtype: array_like
        """

        # Derivatives of the residuals_LM objective function, PSPL version

        lightcurve = telescope.lightcurve_flux

        time = lightcurve[:, 0]
        errflux = lightcurve[:, 2]

        # Derivative of A = (u^2+2)/(u(u^2+4)^0.5). Amplification[0] is A(t).
        # Amplification[1] is U(t).
        Amplification = self.Jacobian_model_magnification(telescope, pyLIMA_parameters)
        dAmplificationdU = (-8) / (Amplification[1] ** 2 * (Amplification[1] ** 2 + 4) ** 1.5)

        # Derivative of U = (uo^2+(t-to)^2/tE^2)^0.5
        dUdto = -(time - pyLIMA_parameters.to) / \
                (pyLIMA_parameters.tE ** 2 * Amplification[1])
        dUduo = pyLIMA_parameters.uo / Amplification[1]
        dUdtE = -(time - pyLIMA_parameters.to) ** 2 / \
                (pyLIMA_parameters.tE ** 3 * Amplification[1])

        # Derivative of the model

        dresdto = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dAmplificationdU * dUdto / errflux
        dresduo = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dAmplificationdU * dUduo / errflux
        dresdtE = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dAmplificationdU * dUdtE / errflux

        if self.blend_flux_ratio==True:
            dresdfs = (Amplification[0] + getattr(pyLIMA_parameters, 'g_' + telescope.name)) / errflux
            dresdg = getattr(pyLIMA_parameters, 'fs_' + telescope.name) / errflux
        else:
            dresdfs = (Amplification[0]) / errflux
            dresdg = 1 / errflux

        jacobi = np.array([dresdto, dresduo, dresdtE, dresdfs, dresdg])

        return jacobi


class ModelFSPL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns: FSPL
        :rtype: string
        """

        return 'FSPL'

    def paczynski_model_parameters(self):
        """ Define the FSPL standard parameters, [to,uo,tE,rho]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'rho': 3}

        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a FSPL model. More details in microlmagnification module.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification, impact_parameter
        :rtype: array_like,array_like
        """

        source_trajectory_x, source_trajectory_y = self.source_trajectory(telescope, pyLIMA_parameters.to,
                                                                          pyLIMA_parameters.uo,
                                                                          pyLIMA_parameters.tE,
                                                                          pyLIMA_parameters)
        rho = pyLIMA_parameters.rho
        gamma = telescope.gamma

        return microlmagnification.amplification_FSPL(source_trajectory_x, source_trajectory_y, rho,
                                                      gamma, self.yoo_table)

    def Jacobian_model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a FSPL model. More details in microlmagnification module.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification, impact_parameter
        :rtype: array_like,array_like
        """

        source_trajectory_x, source_trajectory_y = self.source_trajectory(telescope, pyLIMA_parameters.to,
                                                                          pyLIMA_parameters.uo,
                                                                          pyLIMA_parameters.tE,
                                                                          pyLIMA_parameters)
        rho = pyLIMA_parameters.rho
        gamma = telescope.gamma

        return microlmagnification.Jacobian_amplification_FSPL(source_trajectory_x, source_trajectory_y, rho,
                                                               gamma, self.yoo_table)

    def model_Jacobian(self, telescope, pyLIMA_parameters):
        """ The derivative of a FSPL model

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: jacobi
        :rtype: array_like
        """
        # Derivatives of the residuals_LM objective function, FSPL version

        fake_model = ModelPSPL(self.event)
        fake_model.define_model_parameters()
        lightcurve = telescope.lightcurve_flux
        time = lightcurve[:, 0]
        errflux = lightcurve[:, 2]
        gamma = telescope.gamma

        # Derivative of A = Yoo et al (2004) method.
        Amplification_PSPL = fake_model.Jacobian_model_magnification(telescope, pyLIMA_parameters)

        dAmplification_PSPLdU = (-8) / (Amplification_PSPL[1] ** 2 * (Amplification_PSPL[1] ** 2 + 4) ** (1.5))

        # z_yoo=U/rho
        z_yoo = Amplification_PSPL[1] / pyLIMA_parameters.rho

        dadu = np.zeros(len(Amplification_PSPL[0]))
        dadrho = np.zeros(len(Amplification_PSPL[0]))

        # Far from the lens (z_yoo>>1), then PSPL.
        ind = np.where((z_yoo > self.yoo_table[0][-1]))[0]
        dadu[ind] = dAmplification_PSPLdU[ind]
        dadrho[ind] = -0.0

        # Very close to the lens (z_yoo<<1), then Witt&Mao limit.
        ind = np.where((z_yoo < self.yoo_table[0][0]))[0]
        dadu[ind] = dAmplification_PSPLdU[ind] * (2 * z_yoo[ind] - gamma * (2 - 3 * np.pi / 4) * z_yoo[ind])

        dadrho[ind] = -Amplification_PSPL[0][ind] * Amplification_PSPL[1][ind] / pyLIMA_parameters.rho ** 2 * \
                      (2 - gamma * (2 - 3 * np.pi / 4))

        # FSPL regime (z_yoo~1), then Yoo et al derivatives
        ind = np.where((z_yoo <= self.yoo_table[0][-1]) & (z_yoo >= self.yoo_table[0][0]))[0]

        dadu[ind] = dAmplification_PSPLdU[ind] * (self.yoo_table[1](z_yoo[ind]) - \
                                                  gamma * self.yoo_table[2](z_yoo[ind])) + \
                    Amplification_PSPL[0][ind] * \
                    (self.yoo_table[3](z_yoo[ind]) - \
                     gamma * self.yoo_table[4](z_yoo[ind])) * 1 / pyLIMA_parameters.rho

        dadrho[ind] = -Amplification_PSPL[0][ind] * Amplification_PSPL[1][ind] / pyLIMA_parameters.rho ** 2 * \
                      (self.yoo_table[3](z_yoo[ind]) - gamma * self.yoo_table[4](z_yoo[ind]))

        dUdto = -(time - pyLIMA_parameters.to) / (pyLIMA_parameters.tE ** 2 * Amplification_PSPL[1])

        dUduo = pyLIMA_parameters.uo / Amplification_PSPL[1]

        dUdtE = -(time - pyLIMA_parameters.to) ** 2 / (pyLIMA_parameters.tE ** 3 * Amplification_PSPL[1])

        # Derivative of the model
        dresdto = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dadu * dUdto / errflux

        dresduo = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dadu * dUduo / errflux

        dresdtE = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dadu * dUdtE / errflux

        dresdrho = getattr(pyLIMA_parameters, 'fs_' + telescope.name) * dadrho / errflux

        Amplification_FSPL = self.model_magnification(telescope, pyLIMA_parameters)

        if self.blend_flux_ratio == True:
            dresdfs = (Amplification_FSPL + getattr(pyLIMA_parameters, 'g_' + telescope.name)) / errflux
            dresdg = getattr(pyLIMA_parameters, 'fs_' + telescope.name) / errflux
        else:
            dresdfs = (Amplification_FSPL) / errflux
            dresdg = 1 / errflux

        jacobi = np.array([dresdto, dresduo, dresdtE, dresdrho, dresdfs, dresdg])

        return jacobi


class ModelDSPL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns: DSPL
        :rtype: string
        """
        return 'DSPL'

    def paczynski_model_parameters(self):
        """ Define the DSPL standard parameters, [to1,uo1,to2,uo2,tE,q_F_filters]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        model_dictionary = {'to': 0, 'uo': 1, 'delta_to': 2, 'delta_uo': 3, 'tE': 4}
        filters = [telescope.filter for telescope in self.event.telescopes]

        unique_filters = np.unique(filters)

        for filter in unique_filters:
            model_dictionary['q_flux_' + filter] = len(model_dictionary)

        self.Jacobian_flag = 'No way'
        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a DSPL model.
        From Hwang et al 2013 : http://iopscience.iop.org/article/10.1088/0004-637X/778/1/55/pdf

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification, impact_parameter
        :rtype: array_like,array_like
        """

        source1_trajectory = self.source_trajectory(telescope, pyLIMA_parameters.to, pyLIMA_parameters.uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        to2 = pyLIMA_parameters.to + pyLIMA_parameters.delta_to
        uo2 = pyLIMA_parameters.delta_uo + pyLIMA_parameters.uo
        source2_trajectory = self.source_trajectory(telescope, to2, uo2,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        source1_magnification = microlmagnification.amplification_PSPL(*source1_trajectory)

        source2_magnification = microlmagnification.amplification_PSPL(*source2_trajectory)

        blend_magnification_factor = getattr(pyLIMA_parameters, 'q_flux_' + telescope.filter)

        effective_magnification = (source1_magnification + source2_magnification * blend_magnification_factor) / (
            1 + blend_magnification_factor)

        return effective_magnification


class ModelUSBL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns: USBL
        :rtype: string
        """
        return 'USBL'

    def paczynski_model_parameters(self):
        """ Define the USBL standard parameters, [to,uo,tE,rho, s,q,alpha]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'rho': 3, 'logs': 4, 'logq': 5, 'alpha': 6}

        self.Jacobian_flag = 'No way'
        self.USBL_windows = None
        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a USBL model.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification
        :rtype: array_like
        """

        self.find_origin(pyLIMA_parameters)
        to, uo = self.uo_to_from_uc_tc(pyLIMA_parameters)
        source_trajectoire = self.source_trajectory(telescope, to, uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        magnification = np.zeros(len(source_trajectoire[0]))

        if 'dsdt' in pyLIMA_parameters._fields:

            separation = 10 ** pyLIMA_parameters.logs + \
                         microlorbitalmotion.orbital_motion_2D_separation_shift(self.orbital_motion_model[1],
                                                                                telescope.lightcurve_flux[:, 0],
                                                                                pyLIMA_parameters.dsdt)

        else:

            separation = np.array([10 ** pyLIMA_parameters.logs] * len(source_trajectoire[0]))

        if self.USBL_windows:

            index_USBL = np.where((telescope.lightcurve_flux[:, 0] <= self.USBL_windows[1]) & (
                telescope.lightcurve_flux[:, 0] >= self.USBL_windows[0]))[0]

            Xs = source_trajectoire[0][index_USBL]
            Ys = source_trajectoire[1][index_USBL]

            magnification_USBL = \
                microlmagnification.amplification_USBL(separation[index_USBL], 10 ** pyLIMA_parameters.logq,
                                                       Xs, Ys, pyLIMA_parameters.rho,
                                                       tolerance=0.001)

            magnification[index_USBL] = magnification_USBL

            index_PSBL = np.where((telescope.lightcurve_flux[:, 0] > self.USBL_windows[1]) | (
                telescope.lightcurve_flux[:, 0] < self.USBL_windows[0]))[0]

            magnification_PSBL = microlmagnification.amplification_PSBL(separation[index_PSBL],
                                                                        10 ** pyLIMA_parameters.logq,
                                                                        source_trajectoire[0][index_PSBL],
                                                                        source_trajectoire[1][index_PSBL])

            magnification[index_PSBL] = magnification_PSBL

        else:

            Xs, Ys = source_trajectoire
            magnification = \
                microlmagnification.amplification_USBL(separation, 10 ** pyLIMA_parameters.logq,
                                                       Xs, Ys, pyLIMA_parameters.rho,
                                                       tolerance=0.001)

        return magnification

    def find_caustics(self, separation, mass_ratio):

        caustics, critical_curves = microlcaustics.compute_2_lenses_caustics_points(separation,
                                                                                    mass_ratio,
                                                                                    resolution=100)

        area_of_interest = microlcaustics.find_area_of_interest_around_caustics(caustics, secure_factor=0.1)

        self.caustics = caustics
        self.critical_curves = critical_curves
        self.area_of_interest = area_of_interest

    def find_origin(self, pyLIMA_parameters):

        if self.x_center:
            pass

        else:
            new_origin_x = 0
            new_origin_y = 0

            if self.binary_origin:

                if self.binary_origin == 'central_caustic':
                    new_origin_x, new_origin_y = microlcaustics.change_source_trajectory_center_to_central_caustics_center(
                        10 ** pyLIMA_parameters.logs,
                        10 ** pyLIMA_parameters.logq)

                if self.binary_origin == 'planetary_caustic':

                    new_origin_x, new_origin_y = microlcaustics.change_source_trajectory_center_to_planetary_caustics_center(
                        10 ** pyLIMA_parameters.logs,
                        10 ** pyLIMA_parameters.logq)

            self.x_center = new_origin_x
            self.y_center = new_origin_y

    def uo_to_from_uc_tc(self, pyLIMA_parameters):

        new_origin_x = self.x_center
        new_origin_y = self.y_center

        to = pyLIMA_parameters.to - pyLIMA_parameters.tE * (new_origin_x * np.cos(pyLIMA_parameters.alpha) +
                                                            new_origin_y * np.sin(pyLIMA_parameters.alpha))

        uo = pyLIMA_parameters.uo - (new_origin_x * np.sin(pyLIMA_parameters.alpha) -
                                     new_origin_y * np.cos(pyLIMA_parameters.alpha))

        return to, uo

    def uc_tc_from_uo_to(self, pyLIMA_parameters, to, uo):
        new_origin_x = self.x_center
        new_origin_y = self.y_center

        tc = to + pyLIMA_parameters.tE * (new_origin_x * np.cos(pyLIMA_parameters.alpha) +
                                          new_origin_y * np.sin(pyLIMA_parameters.alpha))

        uc = uo + (new_origin_x * np.sin(pyLIMA_parameters.alpha) -
                   new_origin_y * np.cos(pyLIMA_parameters.alpha))

        return tc, uc

    def find_binary_regime(self, pyLIMA_parameters):

        binary_regime = microlcaustics.find_2_lenses_caustic_regime(10 ** pyLIMA_parameters.logs,
                                                                    10 ** pyLIMA_parameters.logq)
        return binary_regime


class ModelPSBL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns:PSBL
        :rtype: string
        """
        return 'PSBL'

    def paczynski_model_parameters(self):
        """ Define the USBL standard parameters, [to,uo,tE,rho, s,q,alpha]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'logs': 3, 'logq': 4, 'alpha': 5}

        self.Jacobian_flag = 'No way'

        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a USBL model.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification, impact_parameter
        :rtype: array_like,array_like
        """

        self.find_origin(pyLIMA_parameters)
        to, uo = self.uo_to_from_uc_tc(pyLIMA_parameters)

        source_trajectoire = self.source_trajectory(telescope, to, uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        if 'dsdt' in pyLIMA_parameters._fields:

            separation = 10 ** pyLIMA_parameters.logs + \
                         microlorbitalmotion.orbital_motion_2D_separation_shift(self.orbital_motion_model[1],
                                                                                telescope.lightcurve_flux[:, 0],
                                                                                pyLIMA_parameters.dsdt)

        else:

            separation = np.array([10 ** pyLIMA_parameters.logs] * len(source_trajectoire[0]))

        Xs, Ys = source_trajectoire
        magnification = \
            microlmagnification.amplification_PSBL(separation, 10 ** pyLIMA_parameters.logq, Xs, Ys)

        return magnification

    def find_caustics(self, separation, mass_ratio):

        caustics, critical_curves = microlcaustics.compute_2_lenses_caustics_points(separation,
                                                                                    mass_ratio,
                                                                                    resolution=100)

        area_of_interest = microlcaustics.find_area_of_interest_around_caustics(caustics, secure_factor=0.05)

        self.caustics = caustics
        self.critical_curves = critical_curves
        self.area_of_interest = area_of_interest

    def find_origin(self, pyLIMA_parameters):

        if self.x_center:
            pass

        else:
            new_origin_x = 0
            new_origin_y = 0

            if self.binary_origin:

                if self.binary_origin == 'central_caustic':
                    new_origin_x, new_origin_y = microlcaustics.change_source_trajectory_center_to_central_caustics_center(
                        10 ** pyLIMA_parameters.logs,
                        10 ** pyLIMA_parameters.logq)

                if self.binary_origin == 'planetary_caustic':

                    new_origin_x, new_origin_y = microlcaustics.change_source_trajectory_center_to_planetary_caustics_center(
                        10 ** pyLIMA_parameters.logs,
                        10 ** pyLIMA_parameters.logq)

            self.x_center = new_origin_x
            self.y_center = new_origin_y

    def uo_to_from_uc_tc(self, pyLIMA_parameters):

        new_origin_x = self.x_center
        new_origin_y = self.y_center

        to = pyLIMA_parameters.to - pyLIMA_parameters.tE * (new_origin_x * np.cos(pyLIMA_parameters.alpha) +
                                                            new_origin_y * np.sin(pyLIMA_parameters.alpha))

        uo = pyLIMA_parameters.uo - (new_origin_x * np.sin(pyLIMA_parameters.alpha) -
                                     new_origin_y * np.cos(pyLIMA_parameters.alpha))

        return to, uo

    def uc_tc_from_uo_to(self, pyLIMA_parameters, to, uo):
        new_origin_x = self.x_center
        new_origin_y = self.y_center

        tc = to + pyLIMA_parameters.tE * (new_origin_x * np.cos(pyLIMA_parameters.alpha) +
                                          new_origin_y * np.sin(pyLIMA_parameters.alpha))

        uc = uo + (new_origin_x * np.sin(pyLIMA_parameters.alpha) -
                   new_origin_y * np.cos(pyLIMA_parameters.alpha))

        return tc, uc

    def find_binary_regime(self, pyLIMA_parameters):

        binary_regime = microlcaustics.find_2_lenses_caustic_regime(10 ** pyLIMA_parameters.logs,
                                                                    10 ** pyLIMA_parameters.logq)
        return binary_regime


class ModelFSBL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns:FSBL
        :rtype: string
        """
        return 'FSBL'

    def paczynski_model_parameters(self):
        """ Define the USBL standard parameters, [to,uo,tE,rho, s,q,alpha]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'rho': 3, 'logs': 4, 'logq': 5, 'alpha': 6}

        self.Jacobian_flag = 'No way'
        self.USBL_windows = None
        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a USBL model.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification
        :rtype: array_like
        """

        linear_limb_darkening = telescope.gamma * 3 / (2 + telescope.gamma)

        self.find_origin(pyLIMA_parameters)
        to, uo = self.uo_to_from_uc_tc(pyLIMA_parameters)
        source_trajectoire = self.source_trajectory(telescope, to, uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        magnification = np.zeros(len(source_trajectoire[0]))

        if 'dsdt' in pyLIMA_parameters._fields:

            separation = 10 ** pyLIMA_parameters.logs + \
                         microlorbitalmotion.orbital_motion_2D_separation_shift(self.orbital_motion_model[1],
                                                                                telescope.lightcurve_flux[:, 0],
                                                                                pyLIMA_parameters.dsdt)

        else:

            separation = np.array([10 ** pyLIMA_parameters.logs] * len(source_trajectoire[0]))

        if self.USBL_windows:

            index_USBL = np.where((telescope.lightcurve_flux[:, 0] <= self.USBL_windows[1]) & (
                telescope.lightcurve_flux[:, 0] >= self.USBL_windows[0]))[0]

            Xs = source_trajectoire[0][index_USBL]
            Ys = source_trajectoire[1][index_USBL]

            magnification_USBL = \
                microlmagnification.amplification_FSBL(separation[index_USBL], 10 ** pyLIMA_parameters.logq,
                                                       Xs, Ys, pyLIMA_parameters.rho, linear_limb_darkening,
                                                       tolerance=0.001)

            magnification[index_USBL] = magnification_USBL

            index_PSBL = np.where((telescope.lightcurve_flux[:, 0] > self.USBL_windows[1]) | (
                telescope.lightcurve_flux[:, 0] < self.USBL_windows[0]))[0]

            magnification_PSBL = microlmagnification.amplification_PSBL(separation[index_PSBL],
                                                                        10 ** pyLIMA_parameters.logq,
                                                                        source_trajectoire[0][index_PSBL],
                                                                        source_trajectoire[1][index_PSBL])

            magnification[index_PSBL] = magnification_PSBL

        else:

            Xs, Ys = source_trajectoire
            magnification = \
                microlmagnification.amplification_FSBL(separation, 10 ** pyLIMA_parameters.logq,
                                                       Xs, Ys, pyLIMA_parameters.rho, linear_limb_darkening,
                                                       tolerance=0.001)

        return magnification

    def find_caustics(self, separation, mass_ratio):

        caustics, critical_curves = microlcaustics.compute_2_lenses_caustics_points(separation,
                                                                                    mass_ratio,
                                                                                    resolution=100)

        area_of_interest = microlcaustics.find_area_of_interest_around_caustics(caustics, secure_factor=0.1)

        self.caustics = caustics
        self.critical_curves = critical_curves
        self.area_of_interest = area_of_interest

    def find_origin(self, pyLIMA_parameters):

        if self.x_center:
            pass

        else:
            new_origin_x = 0
            new_origin_y = 0

            if self.binary_origin:

                if self.binary_origin == 'central_caustic':
                    new_origin_x, new_origin_y = microlcaustics.change_source_trajectory_center_to_central_caustics_center(
                        10 ** pyLIMA_parameters.logs,
                        10 ** pyLIMA_parameters.logq)

                if self.binary_origin == 'planetary_caustic':

                    new_origin_x, new_origin_y = microlcaustics.change_source_trajectory_center_to_planetary_caustics_center(
                        10 ** pyLIMA_parameters.logs,
                        10 ** pyLIMA_parameters.logq)

            self.x_center = new_origin_x
            self.y_center = new_origin_y

    def uo_to_from_uc_tc(self, pyLIMA_parameters):

        new_origin_x = self.x_center
        new_origin_y = self.y_center

        to = pyLIMA_parameters.to - pyLIMA_parameters.tE * (new_origin_x * np.cos(pyLIMA_parameters.alpha) +
                                                            new_origin_y * np.sin(pyLIMA_parameters.alpha))

        uo = pyLIMA_parameters.uo - (new_origin_x * np.sin(pyLIMA_parameters.alpha) -
                                     new_origin_y * np.cos(pyLIMA_parameters.alpha))

        return to, uo

    def uc_tc_from_uo_to(self, pyLIMA_parameters, to, uo):
        new_origin_x = self.x_center
        new_origin_y = self.y_center

        tc = to + pyLIMA_parameters.tE * (new_origin_x * np.cos(pyLIMA_parameters.alpha) +
                                          new_origin_y * np.sin(pyLIMA_parameters.alpha))

        uc = uo + (new_origin_x * np.sin(pyLIMA_parameters.alpha) -
                   new_origin_y * np.cos(pyLIMA_parameters.alpha))

        return tc, uc

    def find_binary_regime(self, pyLIMA_parameters):

        binary_regime = microlcaustics.find_2_lenses_caustic_regime(10 ** pyLIMA_parameters.logs,
                                                                    10 ** pyLIMA_parameters.logq)
        return binary_regime


class ModelRRLyraePL(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns: RRLyraePL
        :rtype: string
        """
        return 'RRLyraePL'

    def define_pyLIMA_standard_parameters(self):
        """ Define the standard pyLIMA parameters dictionnary."""
        self.number_of_harmonics = self.model_arguments[0]
        self.pyLIMA_standards_dictionnary = self.paczynski_model_parameters()
        if self.parallax_model[0] != 'None':
            self.Jacobian_flag = 'No way'
            self.pyLIMA_standards_dictionnary['piEN'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['piEE'] = len(self.pyLIMA_standards_dictionnary)

            self.event.compute_parallax_all_telescopes(self.parallax_model)
        for telescope in self.event.telescopes:
            self.pyLIMA_standards_dictionnary['fs_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['fb_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)

        self.pyLIMA_standards_dictionnary = OrderedDict(
            sorted(self.pyLIMA_standards_dictionnary.items(), key=lambda x: x[1]))

        self.parameters_boundaries = microlguess.differential_evolution_parameters_boundaries(self)

    def paczynski_model_parameters(self):
        """ Define the PSPL standard parameters, [to,uo,tE]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        # model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'period': 3, 'A1': 4, 'A2': 5, 'A3': 6, 'A4': 7, 'A5': 8,
        # 'A6': 9,
        # 'phi_21': 10, 'phi_31': 11, 'phi_41': 12, 'phi_51': 13, 'phi_61': 14}

        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'period': 3}

        filters = [telescope.filter for telescope in self.event.telescopes]

        unique_filters = np.unique(filters)

        for filter in unique_filters:
            # model_dictionary['AO' + '_' + filter] = len(model_dictionary)
            for i in range(self.number_of_harmonics):
                model_dictionary['A' + str(i + 1) + '_' + filter] = len(model_dictionary)
                model_dictionary['phi' + str(i + 1) + '_' + filter] = len(model_dictionary)
                # if filter != 'I':
                # model_dictionary['phib' + str(i + 1) + '_' + filter] = len(model_dictionary)

        self.Jacobian_flag = 'No way'

        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a PSPL model. More details in microlmagnification module.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification
        :rtype: array_like
        """

        source_trajectoire = self.source_trajectory(telescope, pyLIMA_parameters.to, pyLIMA_parameters.uo,
                                                    pyLIMA_parameters.tE, pyLIMA_parameters)

        return microlmagnification.amplification_PSPL(*source_trajectoire)

    def compute_the_microlensing_model(self, telescope, pyLIMA_parameters):
        """ Compute the microlensing model according the injected parameters. T

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :returns: the microlensing model, the source and the blending flux
        :rtype: tuple, tuple of array like
        """
        lightcurve = telescope.lightcurve_flux
        time = lightcurve[:, 0]

        amplification = self.model_magnification(telescope, pyLIMA_parameters)

        pulsations = self.compute_pulsations(time, telescope.filter, pyLIMA_parameters)
        f_source, f_blending = self.derive_telescope_flux(telescope, pyLIMA_parameters, amplification, pulsations)

        microlensing_model = f_source * amplification * pulsations + f_blending

        return microlensing_model, f_source, f_blending

    def derive_telescope_flux(self, telescope, pyLIMA_parameters, amplification, pulsations):
        """
        Compute the source/blending flux

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :param array_like amplification: an array containing the magnification
        :param array_like pulsations: an array containing the RR Lyrae pulsations


        :returns:  the source and the blending flux
        :rtype: tuple
        """
        try:
            # Fluxes parameters are fitted
            f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope.name) / 2
            f_blending = 2 * getattr(pyLIMA_parameters, 'fb_' + telescope.name) / 2

        except TypeError:

            # Fluxes parameters are estimated through np.polyfit


            lightcurve = telescope.lightcurve_flux
            flux = lightcurve[:, 1]
            errflux = lightcurve[:, 2]
            f_source, f_blending = np.polyfit(amplification * pulsations, flux, 1, w=1 / errflux)

        return f_source, f_blending

    def compute_radius(self, Teff, time, telescope_V, pyLIMA_parameters):

        pulsations = self.compute_pulsations(time, telescope_V.filter, pyLIMA_parameters)
        f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope_V.name) / 2

        f_source_V = f_source * pulsations
        V_magnitude = 27.4 - 2.5 * np.log10(f_source_V)

        # radius = (0.636*10**-((V_magnitude-2*2.689)/5)/Teff**2)
        radius = 10 ** (0.2 * (-(V_magnitude - 2.689 + 0.1) - 10 * np.log10(Teff) + 37.35))
        return radius

    def compute_Teff(self, color):

        # Casagrande 2010
        color += -1.250
        theta_eff = 0.4033 + 0.8171 * color - 0.1987 * color ** 2
        Teff = 5040 / theta_eff

        return Teff

    def compute_color(self, time, telescope_V, telescope_I, pyLIMA_parameters):

        pulsations = self.compute_pulsations(time, telescope_V.filter, pyLIMA_parameters)
        f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope_V.name) / 2

        f_source_V = f_source * pulsations

        pulsations = self.compute_pulsations(time, telescope_I.filter, pyLIMA_parameters)

        f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope_I.name) / 2

        f_source_I = f_source * pulsations

        V_magnitude = 27.4 - 2.5 * np.log10(f_source_V)
        I_magnitude = 27.4 - 2.5 * np.log10(f_source_I)

        return V_magnitude - I_magnitude

    def compute_pulsations(self, time, filter, pyLIMA_parameters):

        time = time - 2456425

        pulsations = 0
        period = getattr(pyLIMA_parameters, 'period')
        # factor = 0.0
        # pulsations = getattr(pyLIMA_parameters, 'AO'+'_' + telescope.filter)
        for i in range(self.number_of_harmonics):
            amplitude = getattr(pyLIMA_parameters, 'A' + str(i + 1) + '_' + filter)
            phase = getattr(pyLIMA_parameters, 'phi' + str(i + 1) + '_' + filter)

            pulsations += amplitude * np.cos(2 * np.pi * (i + 1) / period * time + phase)

        pulsations = 10 ** (pulsations / 2.5)

        return pulsations


class ModelRRLyraeFS(MLModel):
    @property
    def model_type(self):
        """ Return the kind of microlensing model.

        :returns: RRLyraeFS
        :rtype: string
        """
        return 'RRLyraeFS'

    def define_pyLIMA_standard_parameters(self):
        """ Define the standard pyLIMA parameters dictionnary."""
        self.number_of_harmonics = self.model_arguments[0]

        self.pyLIMA_standards_dictionnary = self.paczynski_model_parameters()
        if self.parallax_model[0] != 'None':
            self.Jacobian_flag = 'No way'
            self.pyLIMA_standards_dictionnary['piEN'] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['piEE'] = len(self.pyLIMA_standards_dictionnary)

            self.event.compute_parallax_all_telescopes(self.parallax_model)
        for telescope in self.event.telescopes:
            self.pyLIMA_standards_dictionnary['fs_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)
            self.pyLIMA_standards_dictionnary['fb_' + telescope.name] = len(self.pyLIMA_standards_dictionnary)

        self.pyLIMA_standards_dictionnary = OrderedDict(
            sorted(self.pyLIMA_standards_dictionnary.items(), key=lambda x: x[1]))

        self.parameters_boundaries = microlguess.differential_evolution_parameters_boundaries(self)

    def paczynski_model_parameters(self):
        """ Define the PSPL standard parameters, [to,uo,tE]

        :returns: a dictionnary containing the pyLIMA standards
        :rtype: dict
        """
        # model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'period': 3, 'A1': 4, 'A2': 5, 'A3': 6, 'A4': 7, 'A5': 8,
        # 'A6': 9,
        # 'phi_21': 10, 'phi_31': 11, 'phi_41': 12, 'phi_51': 13, 'phi_61': 14}

        model_dictionary = {'to': 0, 'uo': 1, 'tE': 2, 'theta_E': 3, 'period': 4}

        filters = [telescope.filter for telescope in self.event.telescopes]

        unique_filters = np.unique(filters)

        for filter in unique_filters:
            # model_dictionary['AO' + '_' + filter] = len(model_dictionary)
            for i in range(self.number_of_harmonics):
                model_dictionary['A' + str(i + 1) + '_' + filter] = len(model_dictionary)
                model_dictionary['phi' + str(i + 1) + '_' + filter] = len(model_dictionary)
                # if filter != 'I':
                # model_dictionary['phib' + str(i + 1) + '_' + filter] = len(model_dictionary)

        self.Jacobian_flag = 'No way'
        self.lyrae = stars.Star()
        return model_dictionary

    def model_magnification(self, telescope, pyLIMA_parameters):
        """ The magnification associated to a PSPL model. More details in microlmagnification module.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :return: magnification
        :rtype: array_like
        """

        source_trajectory_x, source_trajectory_y = self.source_trajectory(telescope, pyLIMA_parameters.to,
                                                                          pyLIMA_parameters.uo,
                                                                          pyLIMA_parameters.tE, pyLIMA_parameters)

        telescope_V, telescope_I = self.find_telecopes_V_and_I()

        color = self.compute_color(telescope.lightcurve_flux[:, 0], telescope_V, telescope_I, pyLIMA_parameters)
        teff = self.compute_Teff(color)

        gammas = []
        rhos = self.compute_radius(teff, telescope.lightcurve_flux[:, 0], telescope_V, pyLIMA_parameters) * (
            0.00456 * 10 ** 6 / pyLIMA_parameters.theta_E)

        pyLIMA_parameters.rho = rhos
        # import pdb;
        # pdb.set_trace()
        count = 0
        for temperature in teff:
            self.lyrae.Teff = temperature
            self.lyrae.log_g = 2
            gamma = self.find_gamma(telescope, self.lyrae)
            gammas.append(gamma)

            count += 1

        rho = pyLIMA_parameters.rho
        gamma = np.array(gammas)

        return microlmagnification.amplification_FSPL_for_Lyrae(source_trajectory_x, source_trajectory_y, rho,
                                                                gamma, self.yoo_table)

    def find_gamma(self, telescope, star):

        telescope.find_gamma(star)
        gamma = telescope.gamma
        return gamma

    def compute_the_microlensing_model(self, telescope, pyLIMA_parameters):
        """ Compute the microlens model according the injected parameters. This is modified by child submodel sublclass,
        if not the default microlensing model is returned.

        :param object telescope: a telescope object. More details in telescope module.
        :param object pyLIMA_parameters: a namedtuple which contain the parameters
        :returns: the microlensing model
        :rtype: array_like
        """
        lightcurve = telescope.lightcurve_flux
        time = lightcurve[:, 0]

        amplification = self.model_magnification(telescope, pyLIMA_parameters)

        pulsations = self.compute_pulsations(time, telescope.filter, pyLIMA_parameters)
        f_source, f_blending = self.derive_telescope_flux(telescope, pyLIMA_parameters, amplification, pulsations)

        microlensing_model = f_source * amplification * pulsations + f_blending

        return microlensing_model, f_source, f_blending

    def derive_telescope_flux(self, telescope, pyLIMA_parameters, amplification, pulsations):

        try:
            # Fluxes parameters are fitted
            f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope.name) / 2
            f_blending = 2 * getattr(pyLIMA_parameters, 'fb_' + telescope.name) / 2

        except TypeError:

            # Fluxes parameters are estimated through np.polyfit


            lightcurve = telescope.lightcurve_flux
            flux = lightcurve[:, 1]
            errflux = lightcurve[:, 2]
            f_source, f_blending = np.polyfit(amplification * pulsations, flux, 1, w=1 / errflux)

        return f_source, f_blending

    def compute_radius(self, Teff, time, telescope_V, pyLIMA_parameters, magic_table=None):

        pulsations = self.compute_pulsations(time, telescope_V.filter, pyLIMA_parameters)
        f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope_V.name) / 2

        f_source_V = f_source * pulsations
        V_magnitude = 27.4 - 2.5 * np.log10(f_source_V)

        # radius = (0.636*10**-((V_magnitude-2*2.689)/5)/Teff**2)
        if magic_table:
            BC = magic_table(Teff)
        else:
            BC = 0
        radius = 10 ** (0.2 * (-(V_magnitude - 2.689 + BC) - 10 * np.log10(Teff) + 37.35))
        return radius

    def compute_Teff(self, color):

        # Casagrande 2010
        color += -1.250
        theta_eff = 0.4033 + 0.8171 * color - 0.1987 * color ** 2
        Teff = 5040 / theta_eff

        return Teff

    def compute_color(self, time, telescope_V, telescope_I, pyLIMA_parameters):

        pulsations = self.compute_pulsations(time, telescope_V.filter, pyLIMA_parameters)
        f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope_V.name) / 2

        f_source_V = f_source * pulsations

        pulsations = self.compute_pulsations(time, telescope_I.filter, pyLIMA_parameters)

        f_source = 2 * getattr(pyLIMA_parameters, 'fs_' + telescope_I.name) / 2

        f_source_I = f_source * pulsations

        V_magnitude = 27.4 - 2.5 * np.log10(f_source_V)
        I_magnitude = 27.4 - 2.5 * np.log10(f_source_I)

        return V_magnitude - I_magnitude

    def compute_pulsations(self, time, filter, pyLIMA_parameters):

        time = time - 2456425

        pulsations = 0
        period = getattr(pyLIMA_parameters, 'period')
        # factor = 0.0
        # pulsations = getattr(pyLIMA_parameters, 'AO'+'_' + telescope.filter)
        for i in range(self.number_of_harmonics):
            amplitude = getattr(pyLIMA_parameters, 'A' + str(i + 1) + '_' + filter)
            phase = getattr(pyLIMA_parameters, 'phi' + str(i + 1) + '_' + filter)

            pulsations += amplitude * np.cos(2 * np.pi * (i + 1) / period * time + phase)

        pulsations = 10 ** (pulsations / 2.5)

        return pulsations

    def find_telecopes_V_and_I(self):

        telescope_V = [i for i in self.event.telescopes if 'survey2' in i.name][0]
        telescope_I = [i for i in self.event.telescopes if 'survey1' in i.name][0]

        return telescope_V, telescope_I



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 27 16:39:32 2015

@author: ebachelet
"""
from __future__ import division

import numpy as np
import astropy.io.fits as fits

from pyLIMA import microltoolbox
from pyLIMA import microlparallax

# Conventions for magnitude and flux lightcurves for all pyLIMA. If the injected lightcurve format differs, please
# indicate this in the correponding lightcurve_magnitude_dictionnary or lightcurve_flux_dictionnary, see below.
PYLIMA_LIGHTCURVE_MAGNITUDE_CONVENTION = ['time', 'mag', 'err_mag']
PYLIMA_LIGHTCURVE_FLUX_CONVENTION = ['time', 'flux', 'err_flux']


class Telescope(object):
    """
    ######## Telescope module ########

    This class create a telescope object.

    Attributes :

        location : The location of your observatory. Should be "Earth" (default) or "Space".

        altitude : Altitude in meter of the telescope. Default is 0.0 (sea level).

        longitude : Longitude of the telescope in degrees. Default is 0.57.

        latitude : Latitude in degrees. Default is 49.49 .

        gamma : (Microlensing covention) Limb darkening coefficient :math:`\\Gamma` associated to the filter.
                 The classical (Milne definition) linear limb darkening coefficient can be found using:
                 u=(3*gamma)/(2+gamma).
                 Default is 0.0 (i.e uniform source brightness)

        lightcurve_magnitude : a numpy array describing your data in magnitude.

        lightcurve_magnitude_dictionnary : a python dictionnary to transform the lightcurve_magnitude
                                          input to pyLIMA convention.

        lightcurve_flux : a numpy array describing your data in flux.

        lightcurve_flux_dictionnary : a python dictionnary to transform the lightcurve_flux input to pyLIMA convention.

        reference_flux : a float used for the transformation of difference fluxes to real fluxes. Default is 10000.0 .

        deltas_positions : a list containing the position shift of this observatory due to parallax.
                           Default is an empty list. More details in microlparallax.

    :param string name: name of the telescope. Default is 'NDG'

    :param string camera_filter: telescope filter used. Should be a string which follows the convention of :
               " Gravity and limb-darkening coefficients for the Kepler, CoRoT, Spitzer, uvby,
               UBVRIJHK,
               and Sloan photometric systems"
               Claret, A. and Bloemen, S. 2011A&A...529A..75C. For example, 'I' (default) is
               Johnson_Cousins I filter
               and 'i'' is the SDSS-i' Sloan filter.


    :param array_like light_curve_magnitude:  a numpy array with time, magnitude and error in magnitude.
                                              Default is an None.

    :param dict light_curve_magnitude_dictionnary: a python dictionnary that informs your columns convention. Used to
                                                   translate to pyLIMA convention [time,mag,err_mag].
                                                   Default is {'time': 0, 'mag' : 1, 'err_mag' : 2 }


    :param array-like light_curve_flux:  a numpy array with time, flux and error in flux. Default is an None.

    :param dict light_curve_flux_dictionnary: a python dictionnary that informs your columns convention. Used to
                                              translate to pyLIMA convention [time,flux,err_flux].
                                              Default is {'time': 0, 'flux' : 1, 'err_flux' : 2 }

    :param float reference_flux: a float used for the transformation of difference fluxes to real fluxes.
                                 Default is 0.0 .

    :param str clean_the_lightcurve : a string indicated if you want pyLIMA to clean your lightcurves.
                                      Highly recommanded!
    """

    def __init__(self, name='NDG', camera_filter='I', light_curve_magnitude=None,
                 light_curve_magnitude_dictionnary=None,
                 light_curve_flux=None, light_curve_flux_dictionnary=None,
                 reference_flux=0.0, clean_the_lightcurve='Yes',
                 location = 'Earth', spacecraft_name=None):
        """Initialization of the attributes described above."""

        self.name = name
        self.filter = camera_filter  # Claret2011 convention

        if light_curve_magnitude_dictionnary is None:
            light_curve_magnitude_dictionnary = {'time': 0, 'mag': 1, 'err_mag': 2}

        if light_curve_flux_dictionnary is None:
            light_curve_flux_dictionnary = {'time': 0, 'flux': 1, 'err_flux': 2}

        self.lightcurve_magnitude_dictionnary = light_curve_magnitude_dictionnary
        self.lightcurve_flux_dictionnary = light_curve_flux_dictionnary
        self.reference_flux = reference_flux

        self.lightcurve_magnitude = []
        self.lightcurve_flux = []

        if light_curve_magnitude is None:

            pass

        else:

            self.lightcurve_magnitude = light_curve_magnitude
            self.lightcurve_magnitude = self.arrange_the_lightcurve_columns('magnitude')

            if clean_the_lightcurve == 'Yes':
                self.lightcurve_magnitude = self.clean_data_magnitude()

            self.lightcurve_flux = self.lightcurve_in_flux()

            if clean_the_lightcurve == 'Yes':
                self.lightcurve_flux = self.clean_data_flux()

        if light_curve_flux is None:

            pass

        else:

            self.lightcurve_flux = light_curve_flux
            self.lightcurve_flux = self.arrange_the_lightcurve_columns('flux')

            if clean_the_lightcurve == 'Yes':
                self.lightcurve_flux = self.clean_data_flux()

            self.lightcurve_magnitude = self.lightcurve_in_magnitude()

        self.location = location
        self.altitude = 0.0  # meters
        self.longitude = 0.57  # degrees
        self.latitude = 49.49  # degrees
        self.gamma = 0.0  # This mean you will fit uniform source brightness
        self.deltas_positions = []
        self.spacecraft_name = spacecraft_name # give the true name of the satellite, according to JPL horizon
        self.spacecraft_positions = [] #only for space base observatory, should be a list as
                                       # [dates(JD), ra(degree) , dec(degree) , distances(AU) ]
        self.hidden()

    def arrange_the_lightcurve_columns(self, choice):
        """Rearange the lightcurve to the pyLIMA convention.

        :param string choice: 'magnitude' or 'flux'. A string which indicated on which lightcurves apply
                                   the sorting.


        :return: the lightcurve sorted in pyLIMA convention
        :rtype: array_like
        """

        if choice == 'magnitude':

            lightcurve = []
            for good_column in PYLIMA_LIGHTCURVE_MAGNITUDE_CONVENTION:
                lightcurve.append(self.lightcurve_magnitude[:, self.lightcurve_magnitude_dictionnary[good_column]])

            lightcurve = np.array(lightcurve).T
            return lightcurve

        if choice == 'flux':

            lightcurve = []
            for good_column in PYLIMA_LIGHTCURVE_FLUX_CONVENTION:
                lightcurve.append(self.lightcurve_flux[:, self.lightcurve_flux_dictionnary[good_column]])

            lightcurve = np.array(lightcurve).T
            lightcurve[:, 1] = lightcurve[:, 1] + self.reference_flux
            return lightcurve

    def n_data(self, choice='magnitude'):
        """ Return the number of data points in the lightcurve.

        :param string choice: 'magnitude' (default) or 'flux' The unit you want to check data for.

        :return: the size of the corresponding lightcurve
        :rtype: int
        """

        if choice == 'flux':
            return len(self.lightcurve_flux[:, 0])

        if choice == 'magnitude':
            return len(self.lightcurve_magnitude[:, 0])

    def find_gamma(self, star):
        """
        Set the associated :math:`\\Gamma` linear limb-darkening coefficient associated to the filter.


        :param object star: a stars object.

        """

        self.gamma = star.find_gamma(self.filter)

    def compute_parallax(self, event, parallax):
        """ Compute and set the deltas_positions attribute due to the parallax.

        :param object event: a event object. More details in the event module.
        :param list parallax: a list containing the parallax model and to_par. More details in microlparallax module.
        """
        para = microlparallax.MLParallaxes(event, parallax)
        para.parallax_combination(self)
        print('Parallax(' + parallax[0] + ') estimated for the telescope ' + self.name + ': SUCCESS')

    def clean_data_magnitude(self):
        """
        Clean outliers of the telescope for the fits. Points are considered as outliers if they
        are 10 mag brighter
        or fainter than the lightcurve median or if nan appears in any columns or errobar higher
        than a 1 mag.

        :return: the cleaned magnitude lightcurve
        :rtype: array_like
        """

        maximum_accepted_precision = 1.0

        index = np.where((~np.isnan(self.lightcurve_magnitude).any(axis=1)) &
                         (np.abs(self.lightcurve_magnitude[:, 2]) <= maximum_accepted_precision))[0]

        lightcurve = self.lightcurve_magnitude[index]

        index = np.where((np.isnan(self.lightcurve_magnitude).any(axis=1)) &
                         (np.abs(self.lightcurve_magnitude[:, 2]) > maximum_accepted_precision))[0]
        if len(index) != 0:
            self.bad_points_magnitude = index
            print('pyLIMA found some bad points in the telescope ' + self.name + ', you can found these in the ' \
                                                                                 'bad_points_magnitude attribute.')

        return lightcurve

    def clean_data_flux(self):
        """
        Clean outliers of the telescope for the fits. Points are considered as outliers if they
        are 10 mag brighter
        or fainter than the lightcurve median or if nan appears in any columns or errobar higher
        than a 1 mag.

        :return: the cleaned magnitude lightcurve
        :rtype: array_like
        """

        maximum_accepted_precision = 1.0
        flux = self.lightcurve_flux[:, 1]
        error_flux = self.lightcurve_flux[:, 2]
        index = np.where(
            (~np.isnan(self.lightcurve_flux).any(axis=1)) & (np.abs(error_flux / flux) <= maximum_accepted_precision))[
            0]

        lightcurve = self.lightcurve_flux[index]

        index = np.where(
            (np.isnan(self.lightcurve_flux).any(axis=1)) & (np.abs(error_flux / flux) > maximum_accepted_precision))[0]
        if len(index) != 0:
            self.bad_points_flux = index
            print('pyLIMA found some bad points in the telescope ' + self.name + ', you can found these in the ' \
                                                                                 'bad_points_flux attribute.')
        return lightcurve

    def lightcurve_in_flux(self):
        """
        Transform magnitude to flux using m=27.4-2.5*log10(flux) convention. Transform error bar
        accordingly. More details in microltoolbox module.

        :return: the lightcurve in flux, lightcurve_flux.
        :rtype: array_like
        """

        lightcurve = self.lightcurve_magnitude

        time = lightcurve[:, 0]
        mag = lightcurve[:, 1]
        err_mag = lightcurve[:, 2]

        flux = microltoolbox.magnitude_to_flux(mag)
        error_flux = microltoolbox.error_magnitude_to_error_flux(err_mag, flux)
        lightcurve_in_flux = np.array([time, flux, error_flux]).T

        return lightcurve_in_flux

    def lightcurve_in_magnitude(self):
        """
        Transform flux to magnitude using m = 27.4-2.5*log10(flux) convention. Transform error bar
        accordingly. More details in microltoolbox module.

        :return: the lightcurve in magnitude, lightcurve_magnitude.
        :rtype: array_like
        """
        lightcurve = self.lightcurve_flux

        time = lightcurve[:, 0]
        flux = lightcurve[:, 1]
        error_flux = lightcurve[:, 2]

        magnitude = microltoolbox.flux_to_magnitude(flux)
        error_magnitude = microltoolbox.error_flux_to_error_magnitude(error_flux, flux)

        ligthcurve_magnitude = np.array([time, magnitude, error_magnitude]).T

        return ligthcurve_magnitude

    def hidden(self):

        import webbrowser
        controller = webbrowser.get()

        if self.name =='Mexicola':
            controller.open("https://www.youtube.com/watch?v=GcQdU2qA7D4&t=1684s")


import numpy as np


def orbital_motion_2D_trajectory_shift(to_om, time, dalpha_dt):
    """ Compute the trajectory curvature induced by the orbital motion of the lens.

    :param float to_om: the reference time for the orbital motion
    :param array_like time: the time array to compute the trajectory shift
    :param float dalpha_dt: the angle change rate, in radian/yr

    :return: dalpha, the angle shift
    :rtype: array_like
    """

    dalpha = dalpha_dt * (time - to_om)

    return dalpha


def orbital_motion_2D_separation_shift(to_om, time, ds_dt):
    """ Compute the binary separation change induced by the orbital motion of the lens.

    :param float to_om: the reference time for the orbital motion
    :param array_like time: the time array to compute the trajectory shift
    :param float ds_dt: the binary separation change rate, in einstein_ring_unit/yr

    :return: dseparation, the binary separation shift
    :rtype: array_like
    """
    dseparation = ds_dt * (time - to_om)

    return dseparation


def orbital_motion_3D_separation_shift(time, period, ds_dt):
    """ Compute the binary separation change induced by the orbital motion of the lens.

    :param float to_om: the reference time for the orbital motion
    :param array_like time: the time array to compute the trajectory shift
    :param float ds_dt: the binary separation change rate, in einstein_ring_unit/yr

    :return: dseparation, the binary separation shift
    :rtype: array_like
    """
    dseparation = ds_dt * (time - to_om)

    return dseparation


# -*- coding: utf-8 -*-
"""
Created on Wed Oct 28 11:49:44 2015

@author: ebachelet
"""

import telnetlib

import numpy as np
from astropy import constants as astronomical_constants
from scipy import interpolate
import struct

from pyslalib import slalib

TIMEOUT_JPL = 120  # seconds. The time you allow telnetlib to discuss with JPL, see space_parallax.
JPL_TYPICAL_REQUEST_TIME_PER_LINE = 0.002  # seconds.


### Uncomment the following if the spacecraft dataset is huge! and also in optcallback
# MAX_WINDOW_WIDTH = 80 # Max Value: 65535
# MAX_WINDOW_HEIGHT = 65535 # Max Value: 65535

def horizons_obscodes(observatory):
    """Transform observatory names to JPL horizon codes.
    Write by Tim Lister, thanks :)

    :param str observatory: the satellite name you would like to obtain ephemeris. As to be in the dictionnary
           JPL_HORIZONS_ID (exact name matching!).

    :return: the JPL ID of your satellite.
    :rtype: int
    """

    JPL_HORIZONS_ID = {
        'Geocentric': '500',
        'Kepler': '-227',
        'Spitzer': '-79',
        'HST': '-48',
        'Gaia': '-139479',
        'New Horizons': '-98'
    }

    # Check if we were passed the JPL site code directly
    if (observatory in list(JPL_HORIZONS_ID.values())):
        OBSERVATORY_ID = observatory
    else:
        # Lookup observatory name in map, use ELP's code as a default if not found
        OBSERVATORY_ID = JPL_HORIZONS_ID.get(observatory, 'V37')

    return OBSERVATORY_ID


def optcallback(socket, command, option):
    """Write by Tim Lister, thanks :)
    """
    cnum = ord(command)
    onum = ord(option)
    if cnum == telnetlib.WILL:  # and onum == ECHO:
        socket.write(telnetlib.IAC + telnetlib.DONT + onum)
    if cnum == telnetlib.DO and onum == telnetlib.TTYPE:
        socket.write(telnetlib.IAC + telnetlib.WONT + telnetlib.TTYPE)

        ### Uncomment the following if the spacecraft dataset is huge! and also the global variables
        # at the begining of the module
        # width = struct.pack('H', MAX_WINDOW_WIDTH)
        # height = struct.pack('H', MAX_WINDOW_HEIGHT)
        # socket.send(telnetlib.IAC + telnetlib.SB + telnetlib.NAWS + width + height + telnetlib.IAC + telnetlib.SE)


def compute_parallax_curvature(piE, delta_positions):
    """ Compute the curvature induce by the parallax of from
    deltas_positions of a telescope.

    :param array_like piE: the microlensing parallax vector. Have a look :
                           http://adsabs.harvard.edu/abs/2004ApJ...606..319G
    :param array_like delta_positions: the delta_positions of the telescope. More details in microlparallax module.
    :return: delta_tau and delta_u, the shift introduce by parallax
    :rtype: array_like,array_like
    """

    delta_tau = np.dot(piE, delta_positions)
    delta_beta = np.cross(piE, delta_positions.T)

    return -delta_tau, -delta_beta


class MLParallaxes(object):
    """
    ######## Parallax module ########


    This module compute the parallax shifts due to different parallax effects.

    Attributes :

    event : the event object on which you perform the fit on. More details on the event module.

    parallax_model : The parallax effect you want to fit. Have to be a list containing the
    parallax model name
    and the reference time to_par (in JD unit). Example : ['Annual',2457000.0]

    AU : the astronomical unit,  as defined by astropy (in meter)

    speed_of_light : the speed light c,  as defined by astropy (in meter/second)

    Earth_radius : the Earth equatorial radius,  as defined by astropy (in meter)

    target_angles_in_the_sky : a list containing [RA,DEC] of the target in radians unit.

   :param event: the event object on which you perform the fit on. More details on the event module.
   :param parallax_model: The parallax effect you want to fit. Have to be a list containing the
   parallax model name
        and the to_par value. Example : ['Annual',2457000.0]

    """

    def __init__(self, event, parallax_model):
        """Initialization of the attributes described above."""

        self.event = event
        self.parallax_model = parallax_model[0]
        self.to_par = parallax_model[1]
        self.AU = astronomical_constants.au.value
        self.speed_of_light = astronomical_constants.c.value
        self.Earth_radius = astronomical_constants.R_earth.value

        self.target_angles_in_the_sky = [self.event.ra * np.pi / 180, self.event.dec * np.pi / 180]
        self.North_East_vectors_target()

    def North_East_vectors_target(self):
        """This function define the North and East vectors projected on the sky plane
        perpendicular to the line
        of sight (i.e the line define by RA,DEC of the event).
        """
        target_angles_in_the_sky = self.target_angles_in_the_sky
        Target = np.array(
            [np.cos(target_angles_in_the_sky[1]) * np.cos(target_angles_in_the_sky[0]),
             np.cos(target_angles_in_the_sky[1]) * np.sin(target_angles_in_the_sky[0]),
             np.sin(target_angles_in_the_sky[1])])

        self.East = np.array(
            [-np.sin(target_angles_in_the_sky[0]), np.cos(target_angles_in_the_sky[0]), 0.0])
        self.North = np.cross(Target, self.East)

    def HJD_to_JD(self, time_to_transform):
        """Transform the input time from HJD to JD.

        :param array_like time_to_transform: the numpy array containing the time in HJD you want
        to transform in JD.
        :return: the time in JD
        :rtype: array_like
        """
        AU = self.AU
        light_speed = self.speed_of_light

        time_correction = []
        # DTT=[]


        for time in time_to_transform:

            count = 0
            jd = np.copy(time)

            while count < 3:
                Earth_position = slalib.sla_epv(jd)
                Sun_position = -Earth_position[0]

                Sun_angles = slalib.sla_dcc2s(Sun_position)
                target_angles_in_the_sky = self.target_angles_in_the_sky

                Time_correction = np.sqrt(
                    Sun_position[0] ** 2 + Sun_position[1] ** 2 + Sun_position[
                        2] ** 2) * AU / light_speed * (
                                      np.sin(Sun_angles[1]) * np.sin(
                                          target_angles_in_the_sky[1]) + np.cos(
                                          Sun_angles[1]) * np.cos(
                                          target_angles_in_the_sky[1]) * np.cos(
                                          target_angles_in_the_sky[0] - Sun_angles[0])) / (
                                      3600 * 24.0)
                count = count + 1

        # DTT.append(slalib.sla_dtt(jd)/(3600*24))
        time_correction.append(Time_correction)

        JD = time_to_transform + np.array(time_correction)

        return JD

    def parallax_combination(self, telescope):
        """ Compute, and set, the deltas_positions attributes of the telescope object
       inside the list of telescopes. deltas_positions is the offset between the position of the
       observatory at the time t, and the
       center of the Earth at the date to_par. More details on each parallax functions.

       :param object telescope:  a telescope object on which you want to set the deltas_positions
       due to parallax.

       """

        location = telescope.location

        time = telescope.lightcurve_flux[:, 0]
        delta_North = np.array([])
        delta_East = np.array([])

        if location == 'Earth':

            if (self.parallax_model == 'Annual'):
                telescope_positions = self.annual_parallax(time)

                delta_North = np.append(delta_North, telescope_positions[0])
                delta_East = np.append(delta_East, telescope_positions[1])

            if (self.parallax_model == 'Terrestrial'):
                altitude = telescope.altitude
                longitude = telescope.longitude
                latitude = telescope.latitude

                telescope_positions = self.terrestrial_parallax(time, altitude, longitude, latitude)

                delta_North = np.append(delta_North, telescope_positions[0])
                delta_East = np.append(delta_East, telescope_positions[1])

            if (self.parallax_model == 'Full'):
                telescope_positions = self.annual_parallax(time)

                delta_North = np.append(delta_North, telescope_positions[0])
                delta_East = np.append(delta_East, telescope_positions[1])

                altitude = telescope.altitude
                longitude = telescope.longitude
                latitude = telescope.latitude

                telescope_positions = self.terrestrial_parallax(time, altitude, longitude, latitude)

                delta_North += telescope_positions[0]
                delta_East += telescope_positions[1]

        else:

            telescope_positions = self.annual_parallax(time)
            delta_North = np.append(delta_North, telescope_positions[0])
            delta_East = np.append(delta_East, telescope_positions[1])
            name = telescope.spacecraft_name


            telescope_positions = self.space_parallax(time, name, telescope)
            # import pdb;
            # pdb.set_trace()
            delta_North = delta_North + telescope_positions[0]
            delta_East = delta_East + telescope_positions[1]

        deltas_position = np.array([delta_North, delta_East])

        # set the attributes deltas_positions for the telescope object
        telescope.deltas_positions = deltas_position

    def annual_parallax(self, time_to_treat):
        """Compute the position shift due to the Earth movement. Please have a look on :
        "Resolution of the MACHO-LMC-5 Puzzle: The Jerk-Parallax Microlens Degeneracy"
        Gould, Andrew 2004. http://adsabs.harvard.edu/abs/2004ApJ...606..319G

        :param  time_to_treat: a numpy array containing the time where you want to compute this
        effect.
        :return: the shift induce by the Earth motion around the Sun
        :rtype: array_like

        **WARNING** : this is a geocentric point of view.
                      slalib use MJD time definition, which is MJD = JD-2400000.5
        """

        to_par_mjd = self.to_par - 2400000.5
        Earth_position_time_reference = slalib.sla_epv(to_par_mjd)
        Sun_position_time_reference = -Earth_position_time_reference[0]
        Sun_speed_time_reference = -Earth_position_time_reference[1]
        delta_Sun = []

        for time in time_to_treat:
            time_mjd = time - 2400000.5

            Earth_position = slalib.sla_epv(time_mjd)
            Sun_position = -Earth_position[0]
            delta_sun = Sun_position - (time_mjd - to_par_mjd) * Sun_speed_time_reference - Sun_position_time_reference

            delta_Sun.append(delta_sun.tolist())

        delta_Sun = np.array(delta_Sun)

        delta_Sun_projected = np.array(
            [np.dot(delta_Sun, self.North), np.dot(delta_Sun, self.East)])

        return delta_Sun_projected

    def terrestrial_parallax(self, time_to_treat, altitude, longitude, latitude):
        """ Compute the position shift due to the distance of the obervatories from the Earth
        center.
        Please have a look on :
        "Parallax effects in binary microlensing events"
        Hardy, S.J and Walker, M.A. 1995. http://adsabs.harvard.edu/abs/1995MNRAS.276L..79H

        :param  time_to_treat: a numpy array containing the time where you want to compute this
        effect.
        :param altitude: the altitude of the telescope in meter
        :param longitude: the longitude of the telescope in degree
        :param latitude: the latitude of the telescope in degree
        :return: the shift induce by the distance of the telescope to the Earth center.
        :rtype: array_like

        **WARNING** : slalib use MJD time definition, which is MJD = JD-2400000.5
        """

        radius = (self.Earth_radius + altitude) / self.AU
        Longitude = longitude * np.pi / 180.0
        Latitude = latitude * np.pi / 180.0

        delta_telescope = []
        for time in time_to_treat:
            time_mjd = time - 2400000.5
            sideral_time = slalib.sla_gmst(time_mjd)
            telescope_longitude = - Longitude - self.target_angles_in_the_sky[
                0] + sideral_time

            delta_telescope.append(radius * slalib.sla_dcs2c(telescope_longitude, Latitude))

        delta_telescope = np.array(delta_telescope)
        delta_telescope_projected = np.array(
            [np.dot(delta_telescope, self.North), np.dot(delta_telescope, self.East)])
        return delta_telescope_projected

    def space_parallax(self, time_to_treat, satellite_name, telescope):
        """ Compute the position shift due to the distance of the obervatories from the Earth
        center.
        Please have a look on :
        "Parallax effects in binary microlensing events"
        Hardy, S.J and Walker, M.A. 1995. http://adsabs.harvard.edu/abs/1995MNRAS.276L..79H

        :param  time_to_treat: a numpy array containing the time where you want to compute this
        effect.
        :param satellite_name: the name of the observatory. Have to be recognize by JPL HORIZON.
        :return: the shift induce by the distance of the telescope to the Earth center.
        :rtype: array_like
        **WARNING** : slalib use MJD time definition, which is MJD = JD-2400000.5
        """

        tstart = min(time_to_treat) - 1
        tend = max(time_to_treat) + 1
        if len(telescope.spacecraft_positions) != 0:
        # allow to pass the user to give his own ephemeris
            satellite_positions = np.array(telescope.spacecraft_positions)
        else:
            #call JPL!
            satellite_positions = produce_horizons_ephem(satellite_name, tstart, tend, observatory='Geocentric',
                                                     step_size='60m', verbose=False)[1]

        satellite_positions = np.array(satellite_positions)
        dates = satellite_positions[:, 0].astype(float)
        ra = satellite_positions[:, 1].astype(float)
        dec = satellite_positions[:, 2].astype(float)
        distances = satellite_positions[:, 3].astype(float)

        interpolated_ra = interpolate.interp1d(dates, ra)
        interpolated_dec = interpolate.interp1d(dates, dec)
        interpolated_distance = interpolate.interp1d(dates, distances)

        ra_interpolated = interpolated_ra(time_to_treat)
        dec_interpolated = interpolated_dec(time_to_treat)
        distance_interpolated = interpolated_distance(time_to_treat)

        delta_satellite = []
        for index_time in range(len(time_to_treat)):
            delta_satellite.append(distance_interpolated[index_time] * slalib.sla_dcs2c(
                ra_interpolated[index_time] * np.pi / 180,
                dec_interpolated[index_time] * np.pi / 180))

        delta_satellite = np.array(delta_satellite)
        delta_satellite_projected = np.array(
            [np.dot(delta_satellite, self.North), np.dot(delta_satellite, self.East)])

        return delta_satellite_projected


def produce_horizons_ephem(body, start_time, end_time, observatory='ELP', step_size='60m',
                           verbose=False):
    """
    Write by Tim Lister. Thanks for sharing :) Produce RA,DEC and distance from the Geocentric Center.

    """
    # Lookup observatory name
    OBSERVATORY_ID = horizons_obscodes(observatory)
    body = horizons_obscodes(body)
    if (verbose):
        print("Observatory ID= ", OBSERVATORY_ID)

    tstart = 'JD' + str(start_time)

    if (verbose):
        print("tstart = ", tstart)

    tstop = 'JD' + str(end_time)
    # timeout = TIMEOUT_JPL
    expected_number_of_lines = (end_time - start_time) * 24
    timeout = max(JPL_TYPICAL_REQUEST_TIME_PER_LINE * expected_number_of_lines, 5)

    t = telnetlib.Telnet('horizons.jpl.nasa.gov', 6775)
    t.set_option_negotiation_callback(optcallback)
    data = t.read_until('Horizons> '.encode('utf-8'))
    if (verbose):
        print("data = ", data)
        #        print "hex string = %s\n\n" % binascii.hexlify(data)
    while (data.find('Horizons>'.encode('utf-8')) < 0):
        t.write('\n'.encode('utf-8'))
        data = t.read_until('Horizons> '.encode('utf-8'))
        if (verbose):
            print("data = ", data)
    t.write((body + '\n').encode('utf-8'))
    data = t.read_until('Select ... [E]phemeris, [F]tp, [M]ail, [R]edisplay, ?, <cr>: '.encode('utf-8'),
                        timeout)
    if len(data) == 0:
        print('No connection to JPL, sorry :(')
        flag = 'No connection to JPL'
        return flag

    if (verbose):
        print("data = ", data)
    if (data.find('phemeris'.encode('utf-8')) < 0):
        if (data.find('EXACT'.encode('utf-8')) >= 0):
            t.write('\n'.encode('utf-8'))
            data = t.read_until('Select ... [E]phemeris, [F]tp, [M]ail, [R]edisplay, ?, <cr>: '.encode('utf-8'),
                                timeout)
            if (verbose):
                print(data)
            useID = ''
        else:
            # then we have a conflict in the name.
            # e.g. Titan vs. Titania, or Mars vs. Mars Barycenter
            # Try to resolve by forcing an exact match.
            lines = data.split('\n')
            if (verbose):
                print("Multiple entries found, using exact match")
                print("nlines = %d" % (len(lines)))
            firstline = -1
            lastvalidline = -1
            l = 0
            useID = -1
            for line in lines:
                if (verbose):
                    print(line)
                if (line.find('-----') >= 0):
                    if (firstline == -1):
                        firstline = l + 1
                else:
                    tokens = line.split()
                    if (firstline >= 0 and lastvalidline == -1):
                        if (len(tokens) < 2):
                            lastvalidline = l - 1
                        elif (tokens[1] == body and len(tokens) < 3):
                            # The <3 is necessary to filter out entries for a planet's
                            # barycenter
                            useID = int(tokens[0])
                            useBody = tokens[1]
                            if (verbose):
                                print("Use instead the id = %s = %d" % (tokens[0], useID))
                l = l + 1
            if (useID == -1):
                # Try again with only the first letter capitalized, Probably not necessary
                body = str.upper(body[0]) + str.lower(body[1:])
                #          print "Try the exact match search again with body = ", body
                firstline = -1
                lastvalidline = -1
                l = 0
                for line in lines:
                    if (verbose):
                        print(line)
                    if (line.find('-----') >= 0):
                        if (firstline == -1):
                            firstline = l + 1
                    elif (firstline > 0):
                        if (verbose):
                            print("Splitting this line = %s" % (line))
                        tokens = line.split()
                        if (verbose):
                            print("length=%d,  %d tokens found" % (len(line), len(tokens)))
                        if (firstline >= 0 and lastvalidline == -1):
                            if (len(tokens) < 2):
                                # this is the final (i.e. blank) line in the list
                                lastvalidline = l - 1
                            elif (tokens[1] == body):
                                #                  print "%s %s is equal to %s." % (tokens[
                                # 0],tokens[1],body)
                                useID = int(tokens[0])
                                useBody = tokens[1]
                                if (len(tokens) < 3):
                                    if (verbose):
                                        print("Use instead the id = %s = %d" % (
                                            tokens[0], useID))
                                elif (len(tokens[2].split()) < 1):
                                    if (verbose):
                                        print("Use instead the id = ", tokens[0])
                            else:
                                if (verbose):
                                    print("%s %s is not equal to %s." % (
                                        tokens[0], tokens[1], body))
                    l = l + 1
            if (verbose):
                print("line with first possible source = ", firstline)
                print("line with last possible source = ", lastvalidline)
                print("first possible source = ", (lines[firstline].split())[1])
                print("last possible source = ", (lines[lastvalidline].split())[1])
                print("Writing ", useID)
            t.write((str(useID) + '\n').encode('utf-8'))
            data = t.read_until('Select ... [E]phemeris, [F]tp, [M]ail, [R]edisplay, ?, <cr>: '.encode('utf-8'))
            if (verbose):
                print(data)
    else:
        useID = ''

    t.write('e\n'.encode('utf-8'))
    data = t.read_until('Observe, Elements, Vectors  [o,e,v,?] : '.encode('utf-8'))
    if (verbose):
        print(data)
    t.write('o\n'.encode('utf-8'))
    data = t.read_until('Coordinate center [ <id>,coord,geo  ] : '.encode('utf-8'))
    if (verbose):
        print(data)
    t.write(('%s\n' % OBSERVATORY_ID).encode('utf-8'))
    data = t.read_until('[ y/n ] --> '.encode('utf-8'))
    pointer = data.find('----------------'.encode('utf-8'))
    ending = data[pointer:]
    lines = ending.split('\n'.encode('utf-8'))
    try:
        if (verbose):
            print("Parsing line = %s" % (lines))
        tokens = lines[1].split()
    except:
        print("Telescope code unrecognized by JPL.")
        return ([], [], [])

    if (verbose):
        print(data)
    obsname = ''
    for i in range(4, len(tokens)):
        obsname += (tokens[i]).decode('utf-8')
        if (i < len(tokens) + 1):
            obsname += ' '
    print("Confirmed Observatory name = ", obsname)
    if (useID != ''):
        print("Confirmed Target ID = %d = %s" % (useID, useBody))
    t.write('y\n'.encode('utf-8'))
    data = t.read_until('] : '.encode('utf-8'), 1)
    if (verbose):
        print(data)

    t.write((tstart + '\n').encode('utf-8'))
    data = t.read_until('] : '.encode('utf-8'), 1)
    if (verbose):
        print(data)
    t.write((tstop + '\n').encode('utf-8'))
    data = t.read_until(' ? ] : '.encode('utf-8'), timeout)
    if (verbose):
        print(data)
    t.write((step_size + '\n').encode('utf-8'))
    data = t.read_until(', ?] : '.encode('utf-8'), timeout)
    if (verbose):
        print(data)
    if (1 == 1):
        # t.write('n\n1,3,4,9,19,20,23,\nJ2000\n\n\nMIN\nDEG\nYES\n\n\nYES\n\n\n\n\n\n\n\n')
        t.write('n\n1,20,\nJ2000\n\n\JD\nMIN\nDEG\nYES\n\n\nYES\n\n\n\n\n\n\n100000\n\n'.encode('utf-8'))
    else:
        t.write('y\n'.encode('utf-8'))  # accept default output?
        data = t.read_until(', ?] : '.encode('utf-8'))  # ,timeout)
        if (verbose):
            print(data)
        t.write('1,3\n'.encode('utf-8'))

    t.read_until('$$SOE'.encode('utf-8'), timeout)
    data = t.read_until('$$EOE'.encode('utf-8'), timeout)
    if (verbose):
        print(data)

    t.close()
    lines = data.split('\n'.encode('utf-8'))
    horemp = []
    for hor_line in lines:
        if (verbose):
            print("hor_line = ", hor_line)
            print(len(hor_line.split()))
        data_line = True
        # print hor_line


        if (len(hor_line.split()) == 4):

            (time, raDegrees, decDegrees, light_dist) = hor_line.split()

        elif (len(hor_line.split()) == 0 or len(hor_line.split()) == 1):
            data_line = False
        else:
            data_line = False
            print("Wrong number of fields (", len(hor_line.split()), ")")
            print("hor_line=", hor_line)

        if (data_line == True):

            horemp_line = [time, raDegrees, decDegrees, light_dist]
            if (verbose):
                print(horemp_line)
            horemp.append(horemp_line)

            # Construct ephem_info
    ephem_info = {'obj_id': body,
                  'emp_sitecode': OBSERVATORY_ID,
                  'emp_timesys': '(UT)',
                  'emp_rateunits': '"/min'
                  }
    flag = 'Succes connection to JPL'
    print('Successfully ephemeris from JPL!')
    # import pdb;
    # pdb.set_trace()
    return flag, horemp



# -*- coding: utf-8 -*-
"""
Created on Tue Jun 14 15:49:26 2016

@author: ebachelet
"""
import numpy as np


def microlensing_flux_priors(size_dataset, f_source, g_blending):
    # Little prior here, need to be chaneged

    if (f_source < 0) | (g_blending < -1.0):

        prior_flux_impossible = np.inf
        return prior_flux_impossible

    else:

        prior_flux_impossible = 0.0

    prior_flux_negative_blending = np.log(size_dataset) * 1 / (1 + g_blending)

    prior = prior_flux_impossible + prior_flux_negative_blending
    return prior


def microlensing_parameters_limits_priors(parameters, limits):
    for i in range(len(limits)):

        if (parameters[i] > limits[i][1]) | (parameters[i] < limits[i][0]):

            return np.inf

        else:

            pass
    return 42.0
